# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 27 May 2002
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

from enum import Enum
from typing import Tuple

import numpy as np
from numpy.typing import ArrayLike
from scipy.special import ellipk, ellipe, ellipkinc, ellipeinc

from ...constants import RE, MU, ECCEARTHSQRD, SMALL, TWOPI, J2000_UTC, TUSEC


class OrbitType(Enum):
    CIR_EQUATORIAL = 1  # circular equatorial
    CIR_INCLINED = 2  # circular inclined
    EPH_EQUATORIAL = 3  # elliptical, parabolic, hyperbolic equatorial
    EPH_INCLINED = 4  # elliptical, parabolic, hyperbolic inclined


def determine_orbit_type(ecc: float, incl: float, tol: float = SMALL) -> OrbitType:
    """Determine the type of orbit based on eccentricity and inclination.

    Args:
        ecc (float): The eccentricity of the orbit
        incl (float): The inclination of the orbit in radians
        tol (float, optional): Small value for tolerance

    Returns:
        OrbitType: The type of orbit categorized into one of the following:
                   - circular equatorial
                   - circular inclined
                   - elliptical, parabolic, hyperbolic equatorial
                   - elliptical, parabolic, hyperbolic inclined
    """
    if ecc < tol:
        if (incl < tol) or (abs(incl - np.pi) < tol):
            return OrbitType.CIR_EQUATORIAL
        else:
            return OrbitType.CIR_INCLINED
    elif (incl < tol) or (abs(incl - np.pi) < tol):
        return OrbitType.EPH_EQUATORIAL
    else:
        return OrbitType.EPH_INCLINED


def is_equatorial(inc: float) -> bool:
    """Equatorial check for inclinations.

    Args:
        inc (float): Inclination in radians

    Returns:
        (bool): True if the inclination is equatorial
    """
    return inc < SMALL or abs(inc - np.pi) < SMALL


def elliptic12(
    u: float | ArrayLike, m: float | ArrayLike
) -> Tuple[float | np.ndarray, float | np.ndarray, float | np.ndarray]:
    """Computes the incomplete elliptic integrals of the first and second kind as well
    as the Jacobi Zeta function.

    Args:
        u (float or array_like): Phase in radians
        m (float or array_like): Modulus (0 <= m <= 1)

    Returns:
        tuple: (f, e, z)
            f (float or np.ndarray): Incomplete elliptic integral of the first kind
            e (float or np.ndarray): Incomplete elliptic integral of the second kind
            z (float or np.ndarray): Jacobi Zeta function

    Notes:
        - The MATLAB version sets a maximum value for the modulus m to avoid numerical
          issues, which is not needed/implemented here.
    """
    # Check modulus range
    if np.any(np.array(m) < 0) or np.any(np.array(m) > 1):
        raise ValueError("Modulus m must be in the range 0 <= m <= 1.")

    # Compute incomplete elliptic integrals for the phase u and modulus m
    f = ellipkinc(u, m)  # incomplete elliptic integral of the first kind
    e = ellipeinc(u, m)  # incomplete elliptic integral of the second kind

    # Compute complete elliptic integrals for the modulus m
    k_m = ellipk(m)
    e_m = ellipe(m)

    # Jacobi Zeta function
    z = e - (e_m / k_m) * f

    return f, e, z


def inverse_elliptic2(
    e: float | ArrayLike, m: float | ArrayLike, n_iter: int = 4
) -> np.ndarray:
    """Evaluates the inverse incomplete elliptic integral of the second kind.

    This function is adapted from the MATLAB script `inverselliptic2.m` and uses an
    empirical initialization followed by Newton-Raphson refinement to compute the
    inverse elliptic integral.

    References:
        Elliptic Project, 2011

    Attribution:
        This function is translated and adapted from the MATLAB script
        `inverselliptic2.m` located in the `matlab` directory of this repository. The
        original script contains additional references and details.

    Args:
        e (float or array_like): Value of the integral to be inverted
        m (float or array_like): Modulus (0 <= m <= 1)
        n_iter (int, optional): Number of iterations for Newton-Raphson refinement
                                (defaults to 4)

    Returns:
        np.ndarray: The inverse of the incomplete elliptic integral of the second kind
    """
    # Handle scalar broadcasting
    if np.isscalar(m):
        m = np.full_like(e, m)
    if np.isscalar(e):
        e = np.full_like(m, e)

    # Check modulus range
    if np.any(np.array(m) < 0) or np.any(np.array(m) > 1):
        raise ValueError("Modulus m must be in the range 0 <= m <= 1.")

    # Broadcast m and e to the same shape
    e, m = np.broadcast_arrays(e, m)

    # Complete integral initialization
    e1 = ellipk(m)  # only the complete second kind is needed

    # Calculate empirical initialization
    zeta = 1 - e / e1
    mu = 1 - m
    r = np.sqrt(zeta**2 + mu**2)
    theta = np.arctan2(mu, e + np.finfo(float).eps)
    inv_e = np.pi / 2 + np.sqrt(r) * (theta - np.pi / 2)

    # Newton-Raphson refinement
    for _ in range(n_iter):
        e_calculated = ellipeinc(inv_e, m)
        inv_e -= (e_calculated - e) / np.sqrt(1 - m * np.sin(inv_e) ** 2)

    # Return scalar if inputs were scalar
    return inv_e if inv_e.size > 1 else inv_e.item()


def arclength_ellipse(
    a: float, b: float, theta0: float = 0, theta1: float = TWOPI
) -> float:
    """Calculates the arclength of an ellipse using the elliptic integral of the second
    kind.

    References:
        Elliptic Project, 2011
        http://mathworld.wolfram.com/Ellipse.html

    Attribution:
        This function is translated and adapted from the MATLAB script
        `arclength_ellipse.m` located in the `matlab` directory of this repository.
        The original MATLAB script contains additional references and details.

    Args:
        a (float): Semi-major axis length
        b (float): Semi-minor axis length
        theta0 (float): Start angle in radians (defaults to 0)
        theta1 (float): End angle in radians (defaults to 2pi)

    Returns:
        float: Arclength of the ellipse
    """
    # Circle case
    if a == b:
        return a * (theta1 - theta0)

    # Ellipse with semi-minor axis along x-axis
    if a < b:
        m = 1 - (a / b) ** 2
        e1 = ellipeinc(theta1, m)
        e0 = ellipeinc(theta0, m)
        return b * (e1 - e0)

    # Ellipse with semi-major axis along x-axis
    else:
        m_prime = 1 - (b / a) ** 2
        e1 = ellipeinc(np.pi / 2 - theta1, m_prime)
        e0 = ellipeinc(np.pi / 2 - theta0, m_prime)
        return a * (e0 - e1)


def site(latgd: float, lon: float, alt: float) -> Tuple[np.ndarray, np.ndarray]:
    """Finds the position and velocity vectors for a site.

    The answer is returned in the geocentric equatorial (ECEF) coordinate system.
    Note that the velocity is zero because the coordinate system is fixed to the Earth.

    References:
        Vallado: 2022, p. 432-437, Algorithm 51

    Args:
        latgd (float): Geodetic latitude in radians
        lon (float): Longitude of the site in radians
        alt (float): Altitude in km

    Returns:
        tuple: (rsecef, vsecef)
            rsecef (np.ndarray): ECEF site position vector in km
            vsecef (np.ndarray): ECEF site velocity vector in km/s
    """
    # Compute site position vector
    sinlat = np.sin(latgd)
    cearth = RE / np.sqrt(1 - ECCEARTHSQRD * sinlat**2)
    rdel = (cearth + alt) * np.cos(latgd)
    rk = ((1 - ECCEARTHSQRD) * cearth + alt) * sinlat

    rsecef = np.array([rdel * np.cos(lon), rdel * np.sin(lon), rk])

    # Site velocity vector in ECEF frame is zero
    vsecef = np.zeros(3)

    return rsecef, vsecef


def findc2c3(znew: float) -> Tuple[float, float]:
    """Calculates the c2 and c3 functions for the universal variable z.

    References:
        Vallado: 2022, p. 63, Algorithm 1

    Args:
        znew (float): z variable in rad^2

    Returns:
        tuple: (c2new, c3new)
            c2new (float): c2 function value
            c3new (float): c3 function value
    """
    if znew > SMALL:
        sqrtz = np.sqrt(znew)
        c2new = (1 - np.cos(sqrtz)) / znew
        c3new = (sqrtz - np.sin(sqrtz)) / (sqrtz**3)
    elif znew < -SMALL:
        sqrtz = np.sqrt(-znew)
        c2new = (1 - np.cosh(sqrtz)) / znew
        c3new = (np.sinh(sqrtz) - sqrtz) / (sqrtz**3)
    else:
        c2new = 0.5
        c3new = 1 / 6

    return c2new, c3new


def _calc_gmst(jdut1: float) -> float:
    # Calculate GMST
    # TODO: Formally move to a separate utility function?
    ed = jdut1 - J2000_UTC
    gmst = 99.96779469 + 360.985647366286 * ed + 0.29079e-12 * ed * ed  # deg
    return np.mod(np.radians(gmst), TWOPI)


def lon2nu(jdut1: float, lon: float, incl: float, raan: float, argp: float) -> float:
    """Converts the longitude of the ascending node to the true anomaly.

    References:
        Vallado: 2022, p. 112, Eq. 2-103

    Args:
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        lon (float): Longitude of the ascending node in radians
        incl (float): Orbital inclination in radians
        raan (float): Right ascension of the ascending node in radians
        argp (float): Argument of periapsis in radians

    Returns:
        float: True anomaly in radians (0 to 2pi)
    """
    # Calculate GMST
    gmst = _calc_gmst(jdut1)

    # Calculate lambdau
    lambdau = gmst + lon - raan

    # Ensure lambdau is within 0 to 2pi radians
    lambdau = np.mod(lambdau, TWOPI)

    # Calculate argument of latitude
    arglat = np.arctan(np.tan(lambdau) / np.cos(incl))

    # Adjust arglat for quadrants
    if 0.5 * np.pi <= lambdau < 1.5 * np.pi:
        arglat += np.pi

    return np.mod(arglat - argp, TWOPI)


def nu2lon(jdut1: float, nu: float, incl: float, raan: float, argp: float) -> float:
    """Converts the true anomaly to the longitude of the ascending node.

    References:
        Vallado: 2022, p. 112, Eq. 2-103

    Args:
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        nu (float): True anomaly in radians (0 to 2pi)
        incl (float): Orbital inclination in radians
        raan (float): Right ascension of the ascending node in radians
        argp (float): Argument of periapsis in radians

    Returns:
        float: Longitude of the ascending node in radians (0 to 2pi)
    """
    # Calculate GMST
    gmst = _calc_gmst(jdut1)

    # Compute argument of latitude
    arglat = nu + argp

    # Ensure arglat is within the range [0, 2π]
    arglat = np.mod(arglat, TWOPI)

    # Calculate λu (lambdau)
    lambdau = np.arctan(np.tan(arglat) * np.cos(incl))

    # Adjust lambdau based on quadrant
    if 0.5 * np.pi <= arglat < 1.5 * np.pi:
        lambdau += np.pi

    return np.mod(lambdau - gmst + raan, TWOPI)


def gc2gd(latgc: float) -> float:
    """Converts geocentric latitude to geodetic latitude for positions on the surface of
    the Earth.

    References:
        Vallado: 2022, p. 142, Eq. 3-11

    Args:
        latgc (float): Geocentric latitude in radians

    Returns:
        float: Geodetic latitude in radians (-pi/2 to pi/2)
    """
    return np.arctan(np.tan(latgc) / (1 - ECCEARTHSQRD))


def gd2gc(latgd: float) -> float:
    """Converts geodetic latitude to geocentric latitude for positions on the surface of
    the Earth.

    References:
        Vallado: 2022, p. 142, Eq. 3-11

    Args:
        latgd (float): Geodetic latitude in radians

    Returns:
        float: Geocentric latitude in radians (-pi/2 to pi/2)
    """
    return np.arctan((1 - ECCEARTHSQRD) * np.tan(latgd))


def checkhitearth(
    altpad: float, r1: ArrayLike, v1: ArrayLike, r2: ArrayLike, v2: ArrayLike, nrev: int
) -> Tuple[bool, str]:
    """Checks if the trajectory impacts Earth during the transfer.

    References:
        Vallado: 2022, p. 483-485, Algorithm 58

    Args:
        altpad (float): Altitude pad above the Earth's surface in km
        r1 (array_like): Initial position vector in km
        v1 (array_like): Initial velocity vector in km/s
        r2 (array_like): Final position vector in km
        v2 (array_like): Final velocity vector in km/s
        nrev (int): Number of revolutions (0, 1, 2, ...)

    Returns:
        tuple:
            bool: True if Earth is impacted (False otherwise)
            str: Explanation of the impact status
    """
    # Compute magnitudes of position vectors
    magr1 = np.linalg.norm(r1)
    magr2 = np.linalg.norm(r2)

    # Define the padded radius (Earth's radius + altitude pad)
    rpad = RE + altpad

    # Check if the initial or final position vector is below the padded radius
    if magr1 < rpad or magr2 < rpad:
        return True, "Impact at initial/final radii"

    # Compute dot products of position and velocity vectors
    rdotv1, rdotv2 = np.dot(r1, v1), np.dot(r2, v2)

    # Solve for the reciprocal of the semi-major axis (1/a)
    ainv = 2 / magr1 - np.linalg.norm(v1) ** 2 / MU
    a = 1 / ainv

    # Find ecos(e)
    ecosea1, ecosea2 = 1 - magr1 * ainv, 1 - magr2 * ainv

    # Determine the radius of perigee for nrev > 0
    if nrev > 0:
        if a > 0:
            # Elliptical orbit
            esinea1 = rdotv1 / np.sqrt(MU * a)
            ecc = np.sqrt(ecosea1**2 + esinea1**2)
        else:
            # Hyperbolic orbit
            esinea1 = rdotv1 / np.sqrt(MU * abs(-a))
            ecc = np.sqrt(ecosea1**2 - esinea1**2)

        # Check if the radius of perigee is below the padded radius
        rp = a * (1 - ecc)
        if rp < rpad:
            return True, "Impact during nrev"

    # Check for special cases when nrev = 0
    else:
        if (
            (rdotv1 < 0 < rdotv2)
            or (rdotv1 > 0 < rdotv2 and ecosea1 < ecosea2)
            or (rdotv1 < 0 > rdotv2 and ecosea1 > ecosea2)
        ):

            # Check for parabolic impact
            if abs(ainv) <= SMALL:
                hbar = np.cross(r1, v1)
                magh = np.linalg.norm(hbar)
                rp = magh**2 * 0.5 / MU
                if rp < rpad:
                    return True, "Parabolic impact"

            else:
                esinea1 = rdotv1 / np.sqrt(MU * abs(a))
                if ainv > 0:
                    ecc = np.sqrt(ecosea1**2 + esinea1**2)
                else:
                    ecc = np.sqrt(ecosea1**2 - esinea1**2)

                # Check for elliptical impact
                rp = a * (1 - ecc)
                if ecc < 1 and rp < rpad:
                    return True, "Elliptical impact"

                # Check for hyperbolic impact
                elif rdotv1 < 0 < rdotv2 and rp < rpad:
                    return True, "Hyperbolic impact"

    return False, "No impact"


def checkhitearthc(
    altpadc: float,
    r1c: ArrayLike,
    v1c: ArrayLike,
    r2c: ArrayLike,
    v2c: ArrayLike,
    nrev: int,
) -> Tuple[bool, str]:
    """Checks if the trajectory impacts Earth during the transfer.

    References:
        Vallado: 2022, p. 483-485, Algorithm 58

    Args:
        altpadc (float): Altitude pad above the Earth's surface in Earth radii
        r1c (array_like): Initial position vector in Earth radii
        v1c (array_like): Initial velocity vector in Earth radii per TU
        r2c (array_like): Final position vector in Earth radii
        v2c (array_like): Final velocity vector in Earth radii per TU
        nrev (int): Number of revolutions (0, 1, 2, ...)

    Returns:
        tuple:
            bool: True if Earth is impacted (False otherwise)
            str: Explanation of the impact status
    """
    vconv = RE / TUSEC  # km/s to er/tu
    return checkhitearth(
        altpadc * RE, r1c * RE, v1c * vconv, r2c * RE, v2c * vconv, nrev
    )


def findtof(ro: ArrayLike, r: ArrayLike, p: float) -> float:
    """Finds the time of flight for orbital transfer using p-iteration theory.

    References:
        Vallado: 2022, p. 125-129, Algorithm 11

    Args:
        ro (array_like): Interceptor position vector in km
        r (array_like): Target position vector in km
        p (float): Semiparameter in km

    Returns:
        float: Time of flight in seconds
    """
    # Magnitudes of position vectors
    magr, magro = np.linalg.norm(r), np.linalg.norm(ro)

    # Calculate cosine and sine of change in true anomaly
    cosdnu = np.dot(ro, r) / (magro * magr)
    rcrossr = np.cross(ro, r)
    sindnu = np.linalg.norm(rcrossr) / (magro * magr)

    # Intermediate calculations
    k = magro * magr * (1 - cosdnu)
    l_ = magro + magr
    m = magro * magr * (1 + cosdnu)
    a = (m * k * p) / ((2 * m - l_**2) * p**2 + 2 * k * l_ * p - k**2)

    # Compute f and g
    f = 1 - (magr / p) * (1 - cosdnu)
    g = magro * magr * sindnu / np.sqrt(MU * p)
    alpha = 1 / a

    # Find time of flight based on orbit type
    if alpha > SMALL:
        # Elliptical case
        dnu = np.arctan2(sindnu, cosdnu)
        fdot = (
            np.sqrt(MU / p)
            * np.tan(dnu * 0.5)
            * (((1 - cosdnu) / p) - (1 / magro) - (1 / magr))
        )
        cosdeltae = 1 - (magro / a) * (1 - f)
        sindeltae = (-magro * magr * fdot) / np.sqrt(MU * a)
        deltae = np.arctan2(sindeltae, cosdeltae)
        tof = g + np.sqrt(a**3 / MU) * (deltae - sindeltae)
    elif alpha < SMALL:
        # Hyperbolic case
        deltah = np.arccosh(1 - (magro / a) * (1 - f))
        tof = g + np.sqrt(-(a**3) / MU) * (np.sinh(deltah) - deltah)
    else:
        # Parabolic case
        c = np.sqrt(magr**2 + magro**2 - 2 * magr * magro * cosdnu)
        s = (magro + magr + c) * 0.5
        tof = (2 / 3) * np.sqrt((s**3) * 0.5 / MU) * (1 - ((s - c) / s) ** 1.5)

    return tof



# --------------------------------------------------------------------------------------
# Authors: David Vallado
# Date: 21 July 2006
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

from dataclasses import dataclass
from enum import Enum

import numpy as np
from pydantic import BaseModel

from ... import constants as const


class Classification(Enum):
    # fmt: off
    """Classification of the satellite."""
    Unclassified = "U"
    Classified = "C"


class SatRec(BaseModel):
    a: float = 0.0
    alta: float = 0.0
    altp: float = 0.0
    mo: float = 0.0
    mdot: float = 0.0
    argpo: float = 0.0
    argpdot: float = 0.0
    nodeo: float = 0.0
    nodedot: float = 0.0
    nodecf: float = 0.0
    cc1: float = 0.0
    cc4: float = 0.0
    cc5: float = 0.0
    omgcof: float = 0.0
    xmcof: float = 0.0
    eta: float = 0.0
    sinmao: float = 0.0
    delmo: float = 0.0
    d2: float = 0.0
    d3: float = 0.0
    d4: float = 0.0
    t2cof: float = 0.0
    t3cof: float = 0.0
    t4cof: float = 0.0
    t5cof: float = 0.0
    no: float = 0.0
    ecco: float = 0.0
    inclo: float = 0.0
    isimp: bool = False
    bstar: float = 0.0
    xfact: float = 0.0
    xlamo: float = 0.0
    atime: float = 0.0
    error: Enum = None
    t: float = 0.0
    aycof: float = 0.0
    xlcof: float = 0.0
    x1mth2: float = 0.0
    x7thm1: float = 0.0
    satnum: int = 0
    intldesg: str = ""
    epochyr: int = 0
    epochdays: float = 0.0
    ndot: float = 0.0
    nddot: float = 0.0
    elnum: int = 0
    revnum: int = 0
    no_kozai: float = 0.0
    jdsatepoch: float = 0.0
    jdsatepochf: float = 0.0
    init: bool = True
    classification: Classification = None


class WGSModel(Enum):
    WGS_72_LOW_PRECISION = 721
    WGS_72 = 72
    WGS_84 = 84


@dataclass
class GravitationalConstants:
    tumin: float
    mu: float
    radiusearthkm: float
    xke: float
    j2: float
    j3: float
    j4: float
    j3oj2: float


def getgravc(wgs_model: WGSModel) -> GravitationalConstants:
    """Returns the gravitational constants based on the specified WGS model.

    References:
        - NORAD SpaceTrack Report #3
        - Vallado, Crawford, Hujsak, Kelso, 2006

    Args:
        wgs_model (WGSModel): The WGS model to use

    Returns:
        GravitationalConstants: A data structure containing the gravitational constants
    """
    if wgs_model == WGSModel.WGS_72_LOW_PRECISION:
        mu = 398600.79964
        radiusearthkm = 6378.135
        xke = 0.0743669161
        j2 = 0.001082616
        j3 = -0.00000253881
        j4 = -0.00000165597
    elif wgs_model == WGSModel.WGS_72:
        mu = 398600.8
        radiusearthkm = 6378.135
        xke = const.MIN2SEC / np.sqrt(radiusearthkm**3 / mu)
        j2 = 0.001082616
        j3 = -0.00000253881
        j4 = -0.00000165597
    elif wgs_model == WGSModel.WGS_84:
        mu = 398600.5
        radiusearthkm = 6378.137
        xke = const.MIN2SEC / np.sqrt(radiusearthkm**3 / mu)
        j2 = 0.00108262998905
        j3 = -0.00000253215306
        j4 = -0.00000161098761
    else:
        raise ValueError(f"Unknown option: {wgs_model}")

    return GravitationalConstants(1 / xke, mu, radiusearthkm, xke, j2, j3, j4, j3 / j2)



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 22 January 2018
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

from enum import Enum
from typing import Any, Tuple, Dict

import numpy as np
from scipy.interpolate import CubicSpline

from ... import constants as const
from ...mathtime.julian_date import jday


class JPLInterp(Enum):
    LINEAR = "l"
    SPLINE = "s"


def read_jplde(
    filepath: str, include_hr: bool = True
) -> Tuple[Dict[str, np.ndarray], float, float]:
    """Initializes the JPL planetary ephemeris data by loading the sun and moon
    positions.

    Args:
        filepath (str): Path to the input text file containing ephemeris data
        include_hr (bool, optional): Set to True if the file includes the hour column
                                     (defaults to True)

    Returns:
        tuple: (jpldearr, jdjpldestart, jdjpldestart_frac)
            jpldearr (dict[str, np.ndarray]): Dictionary of JPL DE data records
            jdjpldestart (float): Julian date of the start of the JPL DE data
            jdjpldestart_frac (float): Fractional part of the Julian date at the start
    """
    # Load the input file data
    file_data = np.loadtxt(filepath)

    # Adjust indices based on the presence of the hour column
    offset = 1 if include_hr else 0

    # Extract common fields
    year = file_data[:, 0].astype(int)
    month = file_data[:, 1].astype(int)
    day = file_data[:, 2].astype(int)
    hr = file_data[:, 3] if include_hr else np.zeros_like(year)

    # Construct the JPL DE data dictionary
    jpldearr = {
        "year": year,
        "month": month,
        "day": day,
        "hour": hr,
        "rsun1": file_data[:, 3 + offset],
        "rsun2": file_data[:, 4 + offset],
        "rsun3": file_data[:, 5 + offset],
        "rsmag": file_data[:, 6 + offset],
        "rmoon1": file_data[:, 8 + offset],
        "rmoon2": file_data[:, 9 + offset],
        "rmoon3": file_data[:, 10 + offset],
        "mjd": np.zeros(len(year)),
    }

    # Calculate Modified Julian Date (MJD)
    for i in range(len(year)):
        jd, jd_frac = jday(
            jpldearr["year"][i],
            jpldearr["month"][i],
            jpldearr["day"][i],
            jpldearr["hour"][i],
        )
        jpldearr["mjd"][i] = jd + jd_frac - const.JD_TO_MJD_OFFSET

    # Find the start epoch date
    jdjpldestart, jdjpldestart_frac = jday(
        jpldearr["year"][0],
        jpldearr["month"][0],
        jpldearr["day"][0],
        jpldearr["hour"][0],
    )

    return jpldearr, jdjpldestart, jdjpldestart_frac


def find_jplde_param(
    jdtdb: float,
    jdtdb_f: float,
    jpldearr: dict[str, Any],
    interp: JPLInterp | None = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Finds the JPL DE parameters for a given time using interpolation.

    References:
        Vallado, 2022

    Args:
        jdtdb (float): Epoch Julian date (days from 4713 BC)
        jdtdb_f (float): Fractional part of the epoch Julian date
        jpldearr (dict[str, Any]): Dictionary of JPL DE data records
        interp (JPLInterp, optional): Interpolation method to use (default is None)

    Returns:
        tuple: (rsun, rmoon)
            rsun (np.ndarray): ECI sun position vector in km
            rmoon (np.ndarray): ECI moon position vector in km

    TODO:
        - Fix for the 1/2 day offset case
    """
    # Compute whole-day Julian date and minutes from midnight
    jdb = np.floor(jdtdb + jdtdb_f) + 0.5
    mfme = (jdtdb + jdtdb_f - jdb) * const.DAY2MIN
    if mfme < 0:
        mfme += const.DAY2MIN

    # Determine record index
    jdjpldestarto = np.floor(
        jdtdb + jdtdb_f - jpldearr["mjd"][0] - const.JD_TO_MJD_OFFSET
    )
    if np.any(jpldearr["hour"]):
        # TODO: this works better when the date is closer to the hour, but not so great
        #       when it is close to the 1/2 day; Matlab and C# use a +1 additional
        #       offset (i.e. +2 instead of +1) but this doesn't seem quite right either
        recnum = int(jdjpldestarto) * 2 + 1  # 12-hr data
        if mfme > 720:
            mfme -= 720
    else:
        recnum = int(jdjpldestarto)  # 1-day data

    # Default values if out of bounds
    if not (0 <= recnum <= len(jpldearr["rsun1"]) - 1):
        return np.zeros(3), np.zeros(3)

    # Non-interpolated values
    rsun = np.array(
        [
            jpldearr["rsun1"][recnum],
            jpldearr["rsun2"][recnum],
            jpldearr["rsun3"][recnum],
        ]
    )
    rmoon = np.array(
        [
            jpldearr["rmoon1"][recnum],
            jpldearr["rmoon2"][recnum],
            jpldearr["rmoon3"][recnum],
        ]
    )

    # Fix for interpolation (get back to minutes)
    fixf = mfme / const.DAY2MIN

    # Linear interpolation
    if interp == JPLInterp.LINEAR:
        rsun += fixf * (
            np.array(
                [
                    jpldearr["rsun1"][recnum + 1],
                    jpldearr["rsun2"][recnum + 1],
                    jpldearr["rsun3"][recnum + 1],
                ]
            )
            - rsun
        )
        rmoon += fixf * (
            np.array(
                [
                    jpldearr["rmoon1"][recnum + 1],
                    jpldearr["rmoon2"][recnum + 1],
                    jpldearr["rmoon3"][recnum + 1],
                ]
            )
            - rmoon
        )

    # Cubic spline interpolation
    elif interp == JPLInterp.SPLINE:
        idx1, idx2 = recnum - 1, recnum + 3
        mjds = jpldearr["mjd"][idx1:idx2]

        # Interpolate each component of rsun and rmoon separately
        rsun[0] = CubicSpline(mjds, jpldearr["rsun1"][idx1:idx2])(mjds[1] + fixf)
        rsun[1] = CubicSpline(mjds, jpldearr["rsun2"][idx1:idx2])(mjds[1] + fixf)
        rsun[2] = CubicSpline(mjds, jpldearr["rsun3"][idx1:idx2])(mjds[1] + fixf)
        rmoon[0] = CubicSpline(mjds, jpldearr["rmoon1"][idx1:idx2])(mjds[1] + fixf)
        rmoon[1] = CubicSpline(mjds, jpldearr["rmoon2"][idx1:idx2])(mjds[1] + fixf)
        rmoon[2] = CubicSpline(mjds, jpldearr["rmoon3"][idx1:idx2])(mjds[1] + fixf)

    return rsun, rmoon


def sunmoon(
    jdtdb: float,
    jdtdb_f: float,
    jpldearr: dict[str, Any],
    interp: JPLInterp | None = None,
) -> Tuple[np.ndarray, float, float, np.ndarray, float, float]:
    """Calculates the geocentric equatorial position vectors of the Sun and Moon.

    Args:
        jdtdb (float): Epoch Julian date (days from 4713 BC)
        jdtdb_f (float): Fractional part of the epoch Julian date
        jpldearr (dict[str, Any]): Dictionary of JPL DE data records
        interp (JPLInterp, optional): Interpolation method to use (default is None)

    Returns:
        tuple: (rsun, rtascs, decls, rmoon, rtascm, declm)
            rsun (np.ndarray): ECI sun position vector in km
            rtascs (float): Sun right ascension in radians
            decls (float): Sun declination in radians
            rmoon (np.ndarray): ECI moon position vector in km
            rtascm (float): Moon right ascension in radians
            declm (float): Moon declination in radians
    """
    # Get Sun and Moon position vectors and magnitudes
    rsun, rmoon = find_jplde_param(jdtdb, jdtdb_f, jpldearr, interp)

    # Sun right ascension
    temp_sun = np.hypot(rsun[0], rsun[1])
    rtascs = 0 if temp_sun < const.SMALL else np.arctan2(rsun[1], rsun[0])

    # Sun declination
    decls = np.arcsin(rsun[2] / np.linalg.norm(rsun))

    # Moon right ascension
    temp_moon = np.hypot(rmoon[0], rmoon[1])
    rtascm = 0 if temp_moon < const.SMALL else np.arctan2(rmoon[1], rmoon[0])

    # Moon declination
    declm = np.arcsin(rmoon[2] / np.linalg.norm(rmoon))

    return rsun, rtascs, decls, rmoon, rtascm, declm



# --------------------------------------------------------------------------------------
# Authors: David Vallado, Jeff Beck
# Date: 28 June 2005
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

import logging
import math
from dataclasses import dataclass
from enum import Enum
from typing import Tuple

import numpy as np

from ... import constants as const
from ...mathtime.julian_date import jday
from ...mathtime.calendar import days2mdh
from ..time.sidereal import gstime
from .deep_space import DeepSpace
from .utils import SatRec, WGSModel, getgravc


# Set up logging
logger = logging.getLogger(__name__)

# Constants
JD_EPOCH_1950 = 2433281.5


@dataclass
class SGP4InitOutput:
    ainv: float = 0.0  # may not be used
    ao: float = 0.0
    con41: float = 0.0
    con42: float = 0.0
    cosio: float = 0.0
    cosio2: float = 0.0
    eccsq: float = 0.0
    omeosq: float = 0.0
    posq: float = 0.0
    rp: float = 0.0
    rteosq: float = 0.0
    sinio: float = 0.0
    gsto: float = 0.0
    no_unkozai: float = 0.0


class TypeRun(Enum):
    # fmt: off
    """Character for mode of SGP4 Execution."""
    Catalog = "c"       # +/- 1 day from epoch, 20 min steps
    Verification = "v"  # start/stop/timestep from TLE input (Line 2)
    FromJD = "j"        # start/stop/timestep provided from start and stop Julian dates
    Manual = "m"        # custom start/stop/timestep provided


class Classification(Enum):
    Unclassified = "U"
    Classified = "C"


class PropagationError(Enum):
    INVALID_ELEMENTS = 1
    NEGATIVE_MEAN_MOTION = 2
    ECCENTRICITY_OUT_OF_RANGE = 3
    NEGATIVE_SEMILATUS_RECTUM = 4
    ORBITAL_DECAY = 5


class SGP4:
    """Class for running the SGP4 propagator.

    This class provides the necessary functions to propagate a satellite's orbit using
    the Simplified General Perturbations 4 (SGP4) model.

    References:
        - Hoots, Roehrich, NORAD SpaceTrack Report #3, 1980
        - Hoots, Roehrich, NORAD SpaceTrack Report #6, 1986
        - Hoots, Schumacher, and Glover, 2004
        - Vallado, Crawford, Hujsak, Kelso, 2006

    Args:
        wgs_model (WGSModel): The WGS model to use (default = WGSModel.WGS_84)
        use_afspc_mode (bool): Flag to use AFSPC mode for GST calculation
                               (default = True)

    Attributes:
        wgs_model (WGSModel): The WGS model to use
        use_afspc_mode (bool): Flag to use AFSPC mode for GST calculation
        grav_const (GravitationalConstants): Gravitational constants for the Earth
        satrec (SatRec): Dataclass for satellite elements
        use_deep_space (bool): Flag to use deep space model
        x2o3 (float): 2/3 constant for deep space model
        jdstart_full (float): Start Julian date for TLE start time calculation
        jdstop_full (float): Stop Julian date for TLE stop time calculation
        sgp4init_out (SGP4InitOutput): Output dataclass for SGP4 initialization
        ds (DeepSpace): Deep space object for deep space model

    TODO:
        - This class could be further refactored/cleaned up for better readability.
    """

    def __init__(
        self, wgs_model: WGSModel = WGSModel.WGS_84, use_afspc_mode: bool = True
    ):
        self.wgs_model = wgs_model
        self.use_afspc_mode = use_afspc_mode
        self.grav_const = getgravc(wgs_model)
        self.satrec = SatRec()
        self.use_deep_space = False
        self.x2o3 = 2 / 3

        # TLE attributes
        self.jdstart_full = None
        self.jdstop_full = None

        # SGP4 attributes
        self.sgp4init_out = None

        # Deep space variables
        self.ds = None

    @staticmethod
    def preprocess_tle(tle_line1: str, tle_line2: str) -> Tuple[str, str]:
        # Fix line 1 issues
        tle_line1 = list(tle_line1)
        for j in range(10, 16):
            if tle_line1[j] == " ":
                tle_line1[j] = "_"
        if tle_line1[44] == " ":
            tle_line1[43] = tle_line1[44]
        tle_line1[44] = "."
        if tle_line1[7] == " ":
            tle_line1[7] = "U"
        if tle_line1[9] == " ":
            tle_line1[9] = "."
        for j in range(45, 50):
            if tle_line1[j] == " ":
                tle_line1[j] = "0"
        if tle_line1[51] == " ":
            tle_line1[51] = "0"
        if tle_line1[53] != " ":
            tle_line1[52] = tle_line1[53]
        tle_line1[53] = "."
        if tle_line1[62] == " ":
            tle_line1[62] = "0"
        if len(tle_line1) < 68 or tle_line1[67] == " ":
            tle_line1[67] = "0"

        # Fix line 2 issues
        tle_line2 = list(tle_line2)
        tle_line2[25] = "."
        for j in range(26, 33):
            if tle_line2[j] == " ":
                tle_line2[j] = "0"

        # Convert back to strings
        return "".join(tle_line1), "".join(tle_line2)

    def set_jd_from_from_ymdhms(
        self,
        start_ymdhms: Tuple[int, int, int, int, int, float],
        stop_ymdhms: Tuple[int, int, int, int, int, float],
    ):
        jdstart, jdstartf = jday(*start_ymdhms)
        jdstop, jdstopf = jday(*stop_ymdhms)
        self.jdstart_full = jdstart + jdstartf
        self.jdstop_full = jdstop + jdstopf

    def set_jd_from_yr_doy(
        self, start_yr: int, start_doy: float, stop_yr: int, stop_doy: float
    ):
        start_mdhms = days2mdh(start_yr, start_doy)
        stop_mdhms = days2mdh(stop_yr, stop_doy)

        self.set_jd_from_from_ymdhms((start_yr, *start_mdhms), (stop_yr, *stop_mdhms))

    def twoline2rv(
        self,
        tle_line1: str,
        tle_line2: str,
        typerun: TypeRun = TypeRun.Catalog,
        start: float | None = None,
        stop: float | None = None,
        step: float | None = None,
    ) -> Tuple[float, float, float, np.ndarray, np.ndarray]:
        """Parse TLE lines and populate SGP4 variables.

        This function converts the two line element (TLE) set character string data to
        variables and initializes the sgp4 variables. several intermediate variables
        and quantities are determined. The Verification mode permits quick checks of any
        changes to the underlying technical theory and works using a
        modified tle file in which the start, stop, and delta time values are
        included at the end of the second line of data. The Catalog mode simply
        propagates from -1440 to 1440 min from epoch and is useful when performing
        entire catalog runs.

        If using the FromJD mode, the start and stop Julian dates must be set before
        calling this function (see `set_jd_from_from_ymdhms` or `set_jd_from_yr_doy`).

        Args:
            tle_line1 (str): First line of the TLE set
            tle_line2 (str): Second line of the TLE set
            typerun (TypeRun): Mode of execution (default = TypeRun.Catalog)
            start (float, optional): Start time in minutes from epoch (default = None)
            stop (float, optional): Stop time in minutes from epoch (default = None)
            step (float, optional): Time step in minutes (default = None)

        Returns:
            tuple (r_init, v_init, startmfe, stopmfe, deltamin)
                startmfe (float): Start time in minutes from epoch
                stopmfe (float): Stop time in minutes from epoch
                deltamin (float): Time step in minutes
                r_init (np.ndarray): Initial position vector in TEME frame in km
                v_init (np.ndarray): Initial velocity vector in TEME frame in km/s
        """
        # Constants
        xpdotp = const.DAY2MIN / const.TWOPI  # rev/day / rad/min

        # Preprocess TLE lines
        tle_line1, tle_line2 = self.preprocess_tle(tle_line1, tle_line2)

        # Parse the first line
        self.satrec.satnum = int(tle_line1[2:7])
        self.satrec.classification = Classification(tle_line1[7])
        self.satrec.intldesg = tle_line1[9:17].strip()
        self.satrec.epochyr = int(tle_line1[18:20])
        self.satrec.epochdays = float(tle_line1[20:32])
        self.satrec.ndot = float(tle_line1[33:43])
        self.satrec.nddot = float(tle_line1[44:50]) * 10 ** int(tle_line1[50:52])
        self.satrec.bstar = float(tle_line1[53:59]) * 10 ** int(tle_line1[59:61])
        self.satrec.elnum = int(tle_line1[64:68])

        # Parse the second line
        self.satrec.inclo = np.radians(float(tle_line2[8:16].strip()))
        self.satrec.nodeo = np.radians(float(tle_line2[17:25].strip()))
        self.satrec.ecco = float(f"0.{tle_line2[26:33].strip()}")
        self.satrec.argpo = np.radians(float(tle_line2[34:42].strip()))
        self.satrec.mo = np.radians(float(tle_line2[43:51].strip()))
        self.satrec.no_kozai = float(tle_line2[52:63].strip()) / xpdotp
        self.satrec.revnum = int(tle_line2[63:68].strip())

        # Convert epoch year to full year
        year = self.satrec.epochyr + 2000 if self.satrec.epochyr < 57 else 1900

        # Adjust ndot and nddot units
        self.satrec.ndot /= xpdotp * const.DAY2MIN  # rad/min^2
        self.satrec.nddot /= xpdotp * const.DAY2MIN**2  # rad/min^3

        # Compute Julian date of the epoch
        mdhms = days2mdh(year, self.satrec.epochdays)
        self.satrec.jdsatepoch, self.satrec.jdsatepochf = jday(year, *mdhms)

        # Default values for start, stop, and step
        startmfe, stopmfe, deltamin = 0, const.DAY2MIN, 1

        # Set start, stop, and step based on the type of run
        # Complete catalog evaluation
        if typerun == TypeRun.Catalog:
            startmfe, stopmfe, deltamin = -const.DAY2MIN, const.DAY2MIN, 20

        # Verification - use TLE start/stop/step values
        elif typerun == TypeRun.Verification:
            try:
                startmfe = float(tle_line2[69:81].strip())
                stopmfe = float(tle_line2[82:96].strip())
                deltamin = float(tle_line2[96:105].strip())
            except ValueError:
                raise ValueError("Input TLE does not support verification mode.")

        # From Julian dates (these must be set before calling this function)
        elif typerun == TypeRun.FromJD:
            if any(value is None for value in (self.jdstart_full, self.jdstop_full)):
                raise ValueError(
                    "FromJD mode requires start and stop Julian dates. "
                    "Please set them prior to calling this function!"
                )
            jdsatepoch = self.satrec.jdsatepoch + self.satrec.jdsatepochf
            startmfe = (self.jdstart_full - jdsatepoch) * const.DAY2MIN
            stopmfe = (self.jdstop_full - jdsatepoch) * const.DAY2MIN
            deltamin = step or deltamin

        # Manual mode - use provided start/stop/step values
        elif typerun == TypeRun.Manual:
            startmfe = start or startmfe
            stopmfe = stop or stopmfe
            deltamin = step or deltamin

        # Invalid mode
        else:
            raise ValueError(f"Invalid mode: {typerun}")

        # Initialize SGP4
        epoch = self.satrec.jdsatepoch + self.satrec.jdsatepochf - JD_EPOCH_1950
        r_init, v_init = self.sgp4init(epoch)

        return startmfe, stopmfe, deltamin, r_init, v_init

    def initl(self, epoch: float):
        """Initialize parameters for the SPG4 propagator.

        Args:
            epoch (float): Epoch time in days from Jan 0, 1950, 0 hr

        Returns:
            None
        """
        # Initialize output dataclass
        out = SGP4InitOutput()

        # Calculate auxiliary epoch quantities
        out.eccsq = self.satrec.ecco**2
        out.omeosq = 1 - out.eccsq
        out.rteosq = np.sqrt(out.omeosq)
        out.cosio = np.cos(self.satrec.inclo)
        out.cosio2 = out.cosio**2

        # Un-Kozai the mean motion
        ak = (self.grav_const.xke / self.satrec.no_kozai) ** self.x2o3
        d1 = (
            0.75 * self.grav_const.j2 * (3 * out.cosio2 - 1) / (out.rteosq * out.omeosq)
        )
        delta = d1 / (ak**2)
        adel = ak * (1 - delta**2 - delta * (1 / 3 + 134 * delta**2 / 81))
        delta = d1 / (adel**2)
        out.no_unkozai = self.satrec.no_kozai / (1 + delta)

        # Calculate other terms
        out.ao = (self.grav_const.xke / out.no_unkozai) ** self.x2o3
        out.sinio = np.sin(self.satrec.inclo)
        po = out.ao * out.omeosq
        out.con42 = 1 - 5 * out.cosio2
        out.con41 = -out.con42 - out.cosio2 - out.cosio2
        out.ainv = 1 / out.ao
        out.posq = po**2
        out.rp = out.ao * (1 - self.satrec.ecco)

        # Calculate Greenwich Sidereal Time
        if self.use_afspc_mode:
            out.gsto = gstime(epoch + JD_EPOCH_1950) % const.TWOPI
        else:
            # SGP4 fix - use old way of finding GST
            # Count integer number of days from 0 Jan 1970
            ts70 = epoch - 7305
            ids70 = np.floor(ts70 + 1e-8)
            tfrac = ts70 - ids70
            c1 = 1.72027916940703639e-2
            thgr70 = 1.7321343856509374
            fk5r = 5.07551419432269442e-15
            c1p2p = c1 + const.TWOPI
            out.gsto = np.mod(
                thgr70 + c1 * ids70 + c1p2p * tfrac + ts70 * ts70 * fk5r, const.TWOPI
            )

        self.sgp4init_out = out

    def _adjust_perigee(self, ss, qzms2t):
        """Adjusts sfour and qzms24 for perigees below 156 km."""
        perigee = (self.sgp4init_out.rp - 1) * self.grav_const.radiusearthkm

        # Adjust sfour and qzms24 for perigees below 156 km
        sfour, qzms24 = ss, qzms2t
        if perigee < 156:
            sfour = 20 if perigee < 98 else perigee - 78
            qzms24 = ((120 - sfour) / self.grav_const.radiusearthkm) ** 4
            sfour = sfour / self.grav_const.radiusearthkm + 1

        return sfour, qzms24

    def _compute_perturbation_constants(self, coef, coef1, etasq, eeta, psisq, tsi):
        """Computes perturbation constants for SGP4."""
        cc2 = (
            coef1
            * self.satrec.no
            * (
                self.sgp4init_out.ao * (1 + 1.5 * etasq + eeta * (4 + etasq))
                + 0.375
                * self.grav_const.j2
                * tsi
                / psisq
                * self.sgp4init_out.con41
                * (8 + 3 * etasq * (8 + etasq))
            )
        )
        self.satrec.cc1 = self.satrec.bstar * cc2

        # Compute CC3
        cc3 = 0
        if self.satrec.ecco > 1e-4:
            cc3 = (
                -2
                * coef
                * tsi
                * self.grav_const.j3oj2
                * self.satrec.no
                * self.sgp4init_out.sinio
                / self.satrec.ecco
            )

        # Additional short-periodics parameters
        self.satrec.x1mth2 = 1 - self.sgp4init_out.cosio2
        self.satrec.cc4 = (
            2
            * self.satrec.no
            * coef1
            * self.sgp4init_out.ao
            * self.sgp4init_out.omeosq
            * (
                self.satrec.eta * (2 + 0.5 * etasq)
                + self.satrec.ecco * (0.5 + 2 * etasq)
                - self.grav_const.j2
                * tsi
                / (self.sgp4init_out.ao * psisq)
                * (
                    -3
                    * self.sgp4init_out.con41
                    * (1 - 2 * eeta + etasq * (1.5 - 0.5 * eeta))
                    + 0.75
                    * self.satrec.x1mth2
                    * (2 * etasq - eeta * (1 + etasq))
                    * np.cos(2 * self.satrec.argpo)
                )
            )
        )
        self.satrec.cc5 = (
            2
            * coef1
            * self.sgp4init_out.ao
            * self.sgp4init_out.omeosq
            * (1 + 2.75 * (etasq + eeta) + eeta * etasq)
        )

        return cc3

    def _update_motion_rates(self, pinvsq):
        # Compute higher-order terms
        cosio4 = self.sgp4init_out.cosio2**2
        temp1 = 1.5 * self.grav_const.j2 * pinvsq * self.satrec.no
        temp2 = 0.5 * temp1 * self.grav_const.j2 * pinvsq
        temp3 = -0.46875 * self.grav_const.j4 * pinvsq**2 * self.satrec.no

        # Update motion rates
        self.satrec.mdot = (
            self.satrec.no
            + 0.5 * temp1 * self.sgp4init_out.rteosq * self.sgp4init_out.con41
            + 0.0625
            * temp2
            * self.sgp4init_out.rteosq
            * (13 - 78 * self.sgp4init_out.cosio2 + 137 * cosio4)
        )
        self.satrec.argpdot = (
            -0.5 * temp1 * self.sgp4init_out.con42
            + 0.0625 * temp2 * (7 - 114 * self.sgp4init_out.cosio2 + 395 * cosio4)
            + temp3 * (3 - 36 * self.sgp4init_out.cosio2 + 49 * cosio4)
        )

        xhdot1 = -temp1 * self.sgp4init_out.cosio
        self.satrec.nodedot = (
            xhdot1
            + (
                0.5 * temp2 * (4 - 19 * self.sgp4init_out.cosio2)
                + 2 * temp3 * (3 - 7 * self.sgp4init_out.cosio2)
            )
            * self.sgp4init_out.cosio
        )
        xpidot = self.satrec.argpdot + self.satrec.nodedot

        return xhdot1, xpidot

    def _initialize_deep_space(self, epoch, xpidot):
        # Initialize deep space object
        self.ds = DeepSpace(
            epoch,
            self.satrec.ecco,
            self.satrec.inclo,
            self.satrec.nodeo,
            self.satrec.argpo,
            self.satrec.no,
            self.satrec.mo,
            self.use_afspc_mode,
        )

        # Set attributes and define variables
        self.use_deep_space = True
        self.satrec.isimp = True
        tc = argpm = nodem = mm = 0
        inclm = self.satrec.inclo

        # Call dscom function to compute deep-space common variables
        self.ds.dscom(tc)

        # Call dpper function to adjust for perturbations, if necessary
        if not self.satrec.init:
            self.ds.dpper(self.satrec.t)
            self.satrec.ecco = self.ds.ep
            self.satrec.inclo = self.ds.inclp
            self.satrec.nodeo = self.ds.nodep
            self.satrec.argpo = self.ds.argpp
            self.satrec.mo = self.ds.mp

        # Call dsinit function
        self.ds.dsinit(
            self.satrec,
            self.grav_const.xke,
            tc,
            self.sgp4init_out.gsto,
            xpidot,
            self.sgp4init_out.eccsq,
            inclm,
            nodem,
            argpm,
            mm,
        )

    def _initialize_non_deep_space(self, tsi, sfour):
        cc1sq = self.satrec.cc1**2
        self.satrec.d2 = 4 * self.sgp4init_out.ao * tsi * cc1sq
        temp = self.satrec.d2 * tsi * self.satrec.cc1 / 3
        self.satrec.d3 = (17 * self.sgp4init_out.ao + sfour) * temp
        self.satrec.d4 = (
            0.5
            * temp
            * self.sgp4init_out.ao
            * tsi
            * (221 * self.sgp4init_out.ao + 31 * sfour)
            * self.satrec.cc1
        )
        self.satrec.t3cof = self.satrec.d2 + 2 * cc1sq
        self.satrec.t4cof = 0.25 * (
            3 * self.satrec.d3 + self.satrec.cc1 * (12 * self.satrec.d2 + 10 * cc1sq)
        )
        self.satrec.t5cof = 0.2 * (
            3 * self.satrec.d4
            + 12 * self.satrec.cc1 * self.satrec.d3
            + 6 * self.satrec.d2**2
            + 15 * cc1sq * (2 * self.satrec.d2 + cc1sq)
        )

    def sgp4init(
        self, epoch: float, tol: float = const.SMALL
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Initializes variables for SGP4.

        Args:
            epoch (float): Epoch time in days from Jan 0, 1950 0 hr
            tol (float, optional): Tolerance for small values (default = const.SMALL)

        Returns:
            tuple (r_init, v_init)
                r_init (np.ndarray): Initial position vector in TEME frame in km
                v_init (np.ndarray): Initial velocity vector in TEME frame in km/s

        TODO:
            - Define magic numbers
        """
        # Earth constants
        ss = 78 / self.grav_const.radiusearthkm + 1
        qzms2t = ((120 - 78) / self.grav_const.radiusearthkm) ** 4

        # Initialize SGP4 variables
        self.initl(epoch)

        # Calculate derived orbital parameters
        self.satrec.no = self.sgp4init_out.no_unkozai
        self.satrec.a = (self.satrec.no * self.grav_const.tumin) ** (-2 / 3)
        self.satrec.alta = self.satrec.a * (1 + self.satrec.ecco) - 1
        self.satrec.altp = self.satrec.a * (1 - self.satrec.ecco) - 1

        # Ensure valid orbital elements and positive mean motion
        if not (self.sgp4init_out.omeosq >= 0 and self.satrec.no >= 0):
            r_init, v_init = self.propagate(0)
            return r_init, v_init

        # Determine if perigee is less than 220 km
        if self.sgp4init_out.rp < (220 / self.grav_const.radiusearthkm + 1):
            self.satrec.isimp = True

        # Adjust constants for perigee below 156 km
        sfour, qzms24 = self._adjust_perigee(ss, qzms2t)

        # Definitions
        pinvsq = 1 / self.sgp4init_out.posq
        tsi = 1 / (self.sgp4init_out.ao - sfour)
        self.satrec.eta = self.sgp4init_out.ao * self.satrec.ecco * tsi
        etasq = self.satrec.eta**2
        eeta = self.satrec.ecco * self.satrec.eta
        psisq = abs(1 - etasq)
        coef = qzms24 * tsi**4
        coef1 = coef / psisq**3.5

        # Compute perturbation constants
        cc3 = self._compute_perturbation_constants(coef, coef1, etasq, eeta, psisq, tsi)

        # Compute higher-order terms
        xhdot1, xpidot = self._update_motion_rates(pinvsq)

        # Update other coefficients
        self.satrec.omgcof = self.satrec.bstar * cc3 * np.cos(self.satrec.argpo)
        if self.satrec.ecco > 1e-4:
            self.satrec.xmcof = -self.x2o3 * coef * self.satrec.bstar / eeta
        self.satrec.nodecf = 3.5 * self.sgp4init_out.omeosq * xhdot1 * self.satrec.cc1
        self.satrec.t2cof = 1.5 * self.satrec.cc1

        # Handle divide-by-zero for xinco = 180 degrees
        den = tol
        if abs(self.sgp4init_out.cosio + 1) > tol:
            den = 1 + self.sgp4init_out.cosio

        self.satrec.xlcof = (
            -0.25
            * self.grav_const.j3oj2
            * self.sgp4init_out.sinio
            * (3 + 5 * self.sgp4init_out.cosio)
            / den
        )

        self.satrec.aycof = -0.5 * self.grav_const.j3oj2 * self.sgp4init_out.sinio
        self.satrec.delmo = (1 + self.satrec.eta * np.cos(self.satrec.mo)) ** 3
        self.satrec.sinmao = np.sin(self.satrec.mo)
        self.satrec.x7thm1 = 7 * self.sgp4init_out.cosio2 - 1

        # Determine initialization type
        if (const.TWOPI / self.satrec.no) >= 225:
            # Deep space initialization
            self._initialize_deep_space(epoch, xpidot)
        elif not self.satrec.isimp:
            # Non-deep space initialization
            self._initialize_non_deep_space(tsi, sfour)
        else:
            logger.warning(
                "Neither deep space nor non-deep space coefficients were initialized."
            )

        # Propagate to zero epoch
        r_init, v_init = self.propagate(0)

        return r_init, v_init

    def _apply_secular_gravity_drag(self, t):
        """Apply updates for secular gravity and atmospheric drag."""
        # Initialize secular terms
        t2 = t**2
        xmdf = self.satrec.mo + self.satrec.mdot * t
        argpdf = self.satrec.argpo + self.satrec.argpdot * t
        nodedf = self.satrec.nodeo + self.satrec.nodedot * t
        argpm, mm, nodem = argpdf, xmdf, nodedf + self.satrec.nodecf * t2
        nm, em, inclm = self.satrec.no, self.satrec.ecco, self.satrec.inclo
        tempa = 1 - self.satrec.cc1 * t
        tempe = self.satrec.bstar * self.satrec.cc4 * t
        templ = self.satrec.t2cof * t2

        # Check if non-deep space application
        if not self.satrec.isimp:
            t3, t4 = t**3, t**4
            d2, d3, d4 = self.satrec.d2, self.satrec.d3, self.satrec.d4
            delomg = self.satrec.omgcof * t
            delm = self.satrec.xmcof * (
                (1 + self.satrec.eta * np.cos(xmdf)) ** 3 - self.satrec.delmo
            )
            temp = delomg + delm
            mm, argpm = xmdf + temp, argpdf - temp
            tempa = tempa - d2 * t2 - d3 * t3 - d4 * t4
            tempe = tempe + self.satrec.bstar * self.satrec.cc5 * (
                np.sin(mm) - self.satrec.sinmao
            )
            templ = (
                templ
                + self.satrec.t3cof * t3
                + t4 * (self.satrec.t4cof + t * self.satrec.t5cof)
            )

        # Check if deep space application
        if self.use_deep_space:
            # Update orbital elements
            self.ds.set_mean_elems(em=em, inclm=inclm, nm=nm)

            # Add deep space contributions to mean elements for perturbing 3rd body
            self.ds.dspace(self.satrec, t, self.sgp4init_out.gsto)

            # Unpack updated values directly from `dsinit_out`
            em, inclm, nodem, argpm, nm, mm = self.ds.get_mean_elems()

        # Check if mean motion is less than zero
        if nm <= 0:
            logger.error("Mean motion is less than zero!")
            self.satrec.error = PropagationError.NEGATIVE_MEAN_MOTION

        # Compute some orbital elements
        am = (self.grav_const.xke / nm) ** self.x2o3 * tempa**2
        nm = self.grav_const.xke / am**1.5
        em -= tempe

        # Check if elements are within valid ranges
        if (em >= 1) or (em < -0.001) or (am < 0.95):
            logger.error("Mean elements are out of range!")
            self.satrec.error = PropagationError.INVALID_ELEMENTS

        # Update mean elements
        em = max(em, 1e-6)
        mm += self.satrec.no * templ
        xlm = mm + argpm + nodem
        nodem = math.remainder(nodem, const.TWOPI)
        argpm = math.remainder(argpm, const.TWOPI)
        xlm = math.remainder(xlm, const.TWOPI)
        mm = math.remainder(xlm - argpm - nodem, const.TWOPI)

        return nm, em, inclm, nodem, argpm, mm, am, templ

    def _compute_ds_periodics(self, t, ep, xincp, nodep, argpp, mp, tol):
        """Compute deep space long period periodic contributions"""
        # Set input values
        self.ds.set_attributes(ep=ep, inclp=xincp, nodep=nodep, argpp=argpp, mp=mp)

        # Add deep space periodics
        self.ds.dpper(t)

        # Retrieve updated values
        ep, xincp, nodep, argpp, mp = self.ds.get_attributes(
            "ep", "inclp", "nodep", "argpp", "mp"
        )

        # Handle inclination and node adjustments
        if xincp < 0:
            xincp = -xincp
            nodep += np.pi
            argpp -= np.pi

        # Check eccentricity range
        if (ep < 0) or (ep > 1):
            logger.error(f"Perturbed eccentricity is out of range!: {ep: .2f}")
            self.satrec.error = PropagationError.ECCENTRICITY_OUT_OF_RANGE

        # Compute long period periodics
        sinip, cosip = np.sin(xincp), np.cos(xincp)
        self.satrec.aycof = -0.5 * self.grav_const.j3oj2 * sinip
        den = 1 + cosip if abs(cosip + 1) > tol else const.SMALL
        self.satrec.xlcof = (
            -0.25 * self.grav_const.j3oj2 * sinip * (3 + 5 * cosip) / den
        )

        return ep, xincp, nodep, argpp, mp, sinip, cosip

    def _compute_long_periodics(self, t, em, inclm, nodem, argpm, mm, am, tol):
        """Compute lunar-solar and long period periodics."""
        # Initialize periodics
        ep, xincp, argpp, nodep, mp = em, inclm, argpm, nodem, mm
        sinip, cosip = np.sin(inclm), np.cos(inclm)

        # Update for deep space periodics
        if self.use_deep_space:
            ep, xincp, nodep, argpp, mp, sinip, cosip = self._compute_ds_periodics(
                t, ep, xincp, nodep, argpp, mp, tol
            )

        # Additional periodics
        axnl = ep * np.cos(argpp)
        temp = 1 / (am * (1 - ep**2))
        aynl = ep * np.sin(argpp) + temp * self.satrec.aycof
        xl = mp + argpp + nodep + temp * self.satrec.xlcof * axnl

        return axnl, aynl, xl, xincp, nodep, sinip, cosip

    @staticmethod
    def _solve_keplers_equation(xl, nodep, axnl, aynl, n_iter, tol, lim=0.95):
        """Solve Kepler's equation"""
        u = math.remainder(xl - nodep, const.TWOPI)
        eo1, tem5, ktr = u, np.inf, 1
        sineo1, coseo1 = np.sin(eo1), np.cos(eo1)

        while (abs(tem5) >= tol) and (ktr <= n_iter):
            # Compute correction
            tem5 = 1 - coseo1 * axnl - sineo1 * aynl
            tem5 = (u - aynl * coseo1 + axnl * sineo1 - eo1) / tem5

            # Limit correction to avoid divergence
            if abs(tem5) >= lim:
                tem5 = lim if tem5 > 0 else -lim

            eo1 += tem5
            sineo1, coseo1 = np.sin(eo1), np.cos(eo1)
            ktr += 1

        return sineo1, coseo1

    def _compute_short_periodics(
        self, am, nm, pl, el2, axnl, aynl, coseo1, sineo1, sinip, cosip, nodep, xincp
    ):
        """Compute short period periodics."""
        ecose = axnl * coseo1 + aynl * sineo1
        esine = axnl * sineo1 - aynl * coseo1
        rl = am * (1 - ecose)
        rdotl = np.sqrt(am) * esine / rl
        rvdotl = np.sqrt(pl) / rl
        betal = np.sqrt(1 - el2)
        temp = esine / (1 + betal)
        sinu = am / rl * (sineo1 - aynl - axnl * temp)
        cosu = am / rl * (coseo1 - axnl + aynl * temp)
        su = np.arctan2(sinu, cosu)
        sin2u, cos2u = 2 * sinu * cosu, 1 - 2 * sinu**2
        temp = 1 / pl
        temp1 = 0.5 * self.grav_const.j2 * temp
        temp2 = temp1 * temp

        # Deep space for short period periodics
        if self.use_deep_space:
            cosisq = cosip**2
            self.sgp4init_out.con41 = 3 * cosisq - 1
            self.satrec.x1mth2 = 1 - cosisq
            self.satrec.x7thm1 = 7 * cosisq - 1

        mrt = (
            rl * (1 - 1.5 * temp2 * betal * self.sgp4init_out.con41)
            + 0.5 * temp1 * self.satrec.x1mth2 * cos2u
        )
        su -= 0.25 * temp2 * self.satrec.x7thm1 * sin2u
        xnode = nodep + 1.5 * temp2 * cosip * sin2u
        xinc = xincp + 1.5 * temp2 * cosip * sinip * cos2u
        mvt = rdotl - nm * temp1 * self.satrec.x1mth2 * sin2u / self.grav_const.xke
        rvdot = (
            rvdotl
            + nm
            * temp1
            * (self.satrec.x1mth2 * cos2u + 1.5 * self.sgp4init_out.con41)
            / self.grav_const.xke
        )

        return mrt, su, xnode, xinc, mvt, rvdot

    def propagate(
        self, t: float, n_iter: int = 10, tol: float = const.SMALL
    ) -> Tuple[np.ndarray, np.ndarray]:
        """Simplified General Perturbations 4 (SGP4) model from Space Command.

        This is an updated and combined version of SGP4 and SDP4 algorithms, which
        were originally published separately in SpaceTrack Report #3. This version
        follows the methodology from the AIAA 2006 paper describing the history and
        development of the algorithm.

        Note: `sgp4init` must be called prior to running this function!

        Args:
            t (float): Time since epoch in minutes
            n_iter (int, optional): Number of iterations for solving Kepler's equation
                                    (default = 10)
            tol (float, optional): Tolerance for small values (default = const.SMALL)

        Returns:
            tuple: (r, v)
                r (np.ndarray): Position vector in TEME frame in km
                v (np.ndarray): Velocity vector in TEME frame in km/s

        Return codes for `satrec.error` (non-zero indicates an error)
            1 - Mean elements, ecc >= 1.0 or ecc < -0.001 or a < 0.95 er
            2 - Mean motion < 0.0
            3 - Pert elements, ecc < 0.0  or  ecc > 1.0
            4 - Semi-latus rectum < 0.0
            5 - Satellite has decayed

        TODO:
            - Decide whether to return early if `satrec.error` is non-zero
        """
        # Initialize position and velocity vectors
        r, v = np.zeros(3), np.zeros(3)

        # Compute vkmpersec
        vkmpersec = self.grav_const.radiusearthkm * self.grav_const.xke / const.MIN2SEC

        # Clear error flag and compute time-based quantities
        self.satrec.t, self.satrec.error = t, None

        # Update for secular gravity and atmospheric drag
        nm, em, inclm, nodem, argpm, mm, am, templ = self._apply_secular_gravity_drag(t)

        # Add lunar-solar periodics
        axnl, aynl, xl, xincp, nodep, sinip, cosip = self._compute_long_periodics(
            t, em, inclm, nodem, argpm, mm, am, tol
        )

        # Solve Kepler's equation
        sineo1, coseo1 = self._solve_keplers_equation(
            xl, nodep, axnl, aynl, n_iter, tol
        )

        # Short period preliminary quantities
        el2 = axnl**2 + aynl**2
        pl = am * (1 - el2)

        # Check semi-latus rectum and return if negative
        if pl < 0:
            logger.error("Semi-latus rectum is negative!")
            self.satrec.error = PropagationError.NEGATIVE_SEMILATUS_RECTUM
            return r, v

        # Compute short period periodics
        mrt, su, xnode, xinc, mvt, rvdot = self._compute_short_periodics(
            am, nm, pl, el2, axnl, aynl, coseo1, sineo1, sinip, cosip, nodep, xincp
        )

        # Orientation vectors
        xmx, xmy = -np.sin(xnode) * np.cos(xinc), np.cos(xnode) * np.cos(xinc)
        ux = xmx * np.sin(su) + np.cos(xnode) * np.cos(su)
        uy = xmy * np.sin(su) + np.sin(xnode) * np.cos(su)
        uz = np.sin(xinc) * np.sin(su)
        vx = xmx * np.cos(su) - np.cos(xnode) * np.sin(su)
        vy = xmy * np.cos(su) - np.sin(xnode) * np.sin(su)
        vz = np.sin(xinc) * np.cos(su)

        # Position and velocity vectors
        r = np.array([ux, uy, uz]) * mrt * self.grav_const.radiusearthkm
        v = (np.array([ux, uy, uz]) * mvt + np.array([vx, vy, vz]) * rvdot) * vkmpersec

        # Check for decay condition
        if mrt < 1:
            self.satrec.error = PropagationError.ORBITAL_DECAY

        return r, v



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 16 July 2004
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

from enum import Enum
from typing import Tuple

import numpy as np
from scipy.interpolate import CubicSpline

from .data import IAU06pnOldArray, IAU06Array, IAU06xysArray, EOPArray
from .utils import FundArgs, fundarg, precess
from ... import constants as const
from ...mathtime.julian_date import jday
from ...mathtime.vector import rot1mat, rot2mat, rot3mat


class InterpolationMode(Enum):
    LINEAR = "linear"
    SPLINE = "spline"


def iau06era(jdut1: float) -> np.ndarray:
    """Calculates the transformation matrix that accounts for sidereal time via the
    Earth Rotation Angle (ERA).

    References:
        Vallado, 2022, p. 214

    Args:
        jdut1 (float): Julian date of UT1 (days)

    Returns:
        np.ndarray: 3x3 transformation matrix for PEF to IRE
    """
    # Julian centuries of UT1 (in days from J2000 epoch)
    tut1d = jdut1 - const.J2000

    # Earth rotation angle (ERA) in radians
    era = const.TWOPI * (0.779057273264 + 1.00273781191135448 * tut1d)
    era = np.mod(era, const.TWOPI)

    # Transformation matrix from PEF to IRE
    return np.array(
        [[np.cos(era), -np.sin(era), 0], [np.sin(era), np.cos(era), 0], [0, 0, 1]]
    )


def iau06gst(
    jdut1: float, ttt: float, deltapsi: float, fundargs: FundArgs, iau06arr: IAU06Array
) -> Tuple[float, np.ndarray]:
    """Calculates the IAU 2006 Greenwich Sidereal Time (GST) and transformation matrix.

    References:
        Vallado, 2022, p. 217

    Args:
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        ttt (float): Julian centuries of TT
        deltapsi (float): Change in longitude in radians
        fundargs (FundArgs): Delaunay and planetary arguments
        iau06arr (IAU06Array): IAU 2006 data

    Returns:
        tuple: (gst, st)
            gst (float): Greenwich Sidereal Time in radians (0 to 2pi)
            st (np.ndarray): 3x3 transformation matrix
    """
    # Mean obliquity of the ecliptic
    epsa = (
        84381.406
        - 46.836769 * ttt
        - 0.0001831 * ttt**2
        + 0.0020034 * ttt**3
        - 0.000000576 * ttt**4
        - 0.0000000434 * ttt**5
    )  # arcseconds
    epsa = np.mod(np.radians(epsa / const.DEG2ARCSEC), const.TWOPI)

    # Evaluate the EE complementary terms
    gstsum0, gstsum1 = 0, 0
    n_elem = len(iau06arr.agsti) - 1
    for i in range(n_elem):
        tempval = (
            iau06arr.agsti[i, 0] * fundargs.l
            + iau06arr.agsti[i, 1] * fundargs.l1
            + iau06arr.agsti[i, 2] * fundargs.f
            + iau06arr.agsti[i, 3] * fundargs.d
            + iau06arr.agsti[i, 4] * fundargs.omega
            + iau06arr.agsti[i, 5] * fundargs.lonmer
            + iau06arr.agsti[i, 6] * fundargs.lonven
            + iau06arr.agsti[i, 7] * fundargs.lonear
            + iau06arr.agsti[i, 8] * fundargs.lonmar
            + iau06arr.agsti[i, 9] * fundargs.lonjup
            + iau06arr.agsti[i, 10] * fundargs.lonsat
            + iau06arr.agsti[i, 11] * fundargs.lonurn
            + iau06arr.agsti[i, 12] * fundargs.lonnep
            + iau06arr.agsti[i, 13] * fundargs.precrate
        )
        gstsum0 += iau06arr.agst[i, 0] * np.sin(tempval) + iau06arr.agst[i, 1] * np.cos(
            tempval
        )

    tempval = (
        iau06arr.agsti[n_elem, 0] * fundargs.l
        + iau06arr.agsti[n_elem, 1] * fundargs.l1
        + iau06arr.agsti[n_elem, 2] * fundargs.f
        + iau06arr.agsti[n_elem, 3] * fundargs.d
        + iau06arr.agsti[n_elem, 4] * fundargs.omega
        + iau06arr.agsti[n_elem, 5] * fundargs.lonmer
        + iau06arr.agsti[n_elem, 6] * fundargs.lonven
        + iau06arr.agsti[n_elem, 7] * fundargs.lonear
        + iau06arr.agsti[n_elem, 8] * fundargs.lonmar
        + iau06arr.agsti[n_elem, 9] * fundargs.lonjup
        + iau06arr.agsti[n_elem, 10] * fundargs.lonsat
        + iau06arr.agsti[n_elem, 11] * fundargs.lonurn
        + iau06arr.agsti[n_elem, 12] * fundargs.lonnep
        + iau06arr.agsti[n_elem, 13] * fundargs.precrate
    )
    gstsum1 += iau06arr.agst[n_elem, 0] * ttt * np.sin(tempval) + iau06arr.agst[
        n_elem, 1
    ] * ttt * np.cos(tempval)
    eect2000 = gstsum0 + gstsum1 * ttt

    # Equation of the equinoxes
    ee2000 = deltapsi * np.cos(epsa) + eect2000

    # Earth rotation angle (ERA)
    tut1d = jdut1 - const.J2000  # days from the Jan 1, 2000 12h epoch (UT1)
    era = const.TWOPI * (0.779057273264 + 1.00273781191135448 * tut1d)
    era = np.mod(era, const.TWOPI)

    # Greenwich Mean Sidereal Time (GMST), IAU 2000
    gmst2000 = era + (
        (
            0.014506
            + 4612.156534 * ttt
            + 1.3915817 * ttt**2
            - 0.00000044 * ttt**3
            + 0.000029956 * ttt**4
            + 0.0000000368 * ttt**5
        )
        * const.ARCSEC2RAD
    )

    # Greenwich Sidereal Time (GST)
    gst = gmst2000 + ee2000

    # Transformation matrix
    st = np.array(
        [[np.cos(gst), -np.sin(gst), 0], [np.sin(gst), np.cos(gst), 0], [0, 0, 1]]
    )

    return gst, st


########################################################################################
# IAU 2006 Precession-Nutation Theories (IAU2006/2000A and IAU2006/2000B)
########################################################################################


def _build_transformation_matrices(
    ttt: float, deltaeps: float, deltapsi: float, use_extended_prec: bool
) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
    """Constructs nutation, precession, and combined precession-nutation matrices.

    Args:
        ttt (float): Julian centuries of TT
        deltaeps (float): Nutation in obliquity in radians
        deltapsi (float): Nutation in longitude in radians
        use_extended_prec (bool): Whether to include extended precession terms

    Returns:
        tuple:
            nut (np.ndarray): Nutation matrix (mean to true transformation)
            prec (np.ndarray): Precession matrix (J2000 to date transformation)
            pnb (np.ndarray): Combined precession-nutation matrix (ICRS to GCRF)
    """
    # Get precession angles
    _, psia, wa, ea, xa = precess(ttt, opt="06")

    # Obliquity of the ecliptic
    oblo = 84381.406 * const.ARCSEC2RAD

    # Nutation matrix
    a1 = rot1mat(ea + deltaeps)
    a2 = rot3mat(deltapsi)
    a3 = rot1mat(-ea)
    nut = a3 @ a2 @ a1

    # Precession matrix
    a4 = rot3mat(-xa)
    a5 = rot1mat(wa)
    a6 = rot3mat(psia)
    a7 = rot1mat(-oblo)

    # ICRS to J2000
    a8 = rot1mat(-0.0068192 * const.ARCSEC2RAD)
    a9 = rot2mat(0.041775 * np.sin(oblo) * const.ARCSEC2RAD)
    a10 = rot3mat(0.0146 * const.ARCSEC2RAD)

    # Precession and combined matrices
    if use_extended_prec:
        prec = a10 @ a9 @ a8 @ a7 @ a6 @ a5 @ a4
        pnb = prec @ nut
    else:
        prec = a7 @ a6 @ a5 @ a4
        pnb = a10 @ a9 @ a8 @ prec @ nut

    return nut, prec, pnb


def iau06pna(
    ttt: float, iau06arr: IAU06pnOldArray
) -> Tuple[float, np.ndarray, np.ndarray, np.ndarray, FundArgs]:
    """Calculates the transformation matrix that accounts for the effects of
    precession-nutation using the IAU2006 precession theory and the IAU2000A nutation
    model.

    References:
        Vallado, 2022, p. 214-216

    Args:
        ttt (float): Julian centuries of TT
        iau06arr (IAU06pnOldArray): IAU 2006 data (old nutation coefficients)

    Returns:
        tuple: (deltapsi, pnb, prec, nut, fundargs)
            deltapsi (float): Change in longitude in radians
            pnb (np.ndarray): Combined precession-nutation matrix
            prec (np.ndarray): Precession transformation matrix (MOD to J2000)
            nut (np.ndarray): Nutation transformation matrix (IRE to GCRF)
            fundargs (FundArgs): Delaunay and planetary arguments
    """
    # Obtain data for calculations from the IAU 2006 nutation theory
    fundargs = fundarg(ttt, opt="06")

    # Compute luni-solar nutation
    pnsum, ensum = 0, 0
    for i in range(len(iau06arr.apni) - 1, -1, -1):
        tempval = (
            iau06arr.apni[i, 0] * fundargs.l
            + iau06arr.apni[i, 1] * fundargs.l1
            + iau06arr.apni[i, 2] * fundargs.f
            + iau06arr.apni[i, 3] * fundargs.d
            + iau06arr.apni[i, 4] * fundargs.omega
        )
        tempval = np.mod(tempval, const.TWOPI)
        pnsum += (iau06arr.apn[i, 0] + iau06arr.apn[i, 1] * ttt) * np.sin(
            tempval
        ) + iau06arr.apn[i, 4] * np.cos(tempval)
        ensum += (iau06arr.apn[i, 2] + iau06arr.apn[i, 3] * ttt) * np.cos(
            tempval
        ) + iau06arr.apn[i, 6] * np.sin(tempval)

    # Compute planetary nutation
    pplnsum, eplnsum = 0, 0
    for i in range(len(iau06arr.appli)):
        tempval = (
            iau06arr.appli[i, 0] * fundargs.l
            + iau06arr.appli[i, 1] * fundargs.l1
            + iau06arr.appli[i, 2] * fundargs.f
            + iau06arr.appli[i, 3] * fundargs.d
            + iau06arr.appli[i, 4] * fundargs.omega
            + iau06arr.appli[i, 5] * fundargs.lonmer
            + iau06arr.appli[i, 6] * fundargs.lonven
            + iau06arr.appli[i, 7] * fundargs.lonear
            + iau06arr.appli[i, 8] * fundargs.lonmar
            + iau06arr.appli[i, 9] * fundargs.lonjup
            + iau06arr.appli[i, 10] * fundargs.lonsat
            + iau06arr.appli[i, 11] * fundargs.lonurn
            + iau06arr.appli[i, 12] * fundargs.lonnep
            + iau06arr.appli[i, 13] * fundargs.precrate
        )
        pplnsum += iau06arr.appl[i, 0] * np.sin(tempval) + iau06arr.appl[i, 1] * np.cos(
            tempval
        )
        eplnsum += iau06arr.appl[i, 2] * np.sin(tempval) + iau06arr.appl[i, 3] * np.cos(
            tempval
        )

    # Combine nutation components
    deltapsi = pnsum + pplnsum
    deltaeps = ensum + eplnsum

    # Apply IAU 2006 corrections
    j2d = -2.7774e-6 * ttt * const.ARCSEC2RAD
    deltapsi += deltapsi * (0.4697e-6 + j2d)
    deltaeps += deltaeps * j2d

    # Build transformation matrices
    nut, prec, pnb = _build_transformation_matrices(ttt, deltaeps, deltapsi, False)

    return deltapsi, pnb, prec, nut, fundargs


def iau06pnb(
    ttt: float, iau06arr: IAU06pnOldArray
) -> Tuple[float, np.ndarray, np.ndarray, np.ndarray, FundArgs]:
    """Calculates the transformation matrix that accounts for the effects of
    precession-nutation using the IAU2006 precession theory and a simplified nutation
    model based on IAU2000B.

    References:
        Vallado, 2022, p. 214-216

    Args:
        ttt (float): Julian centuries of TT
        iau06arr (IAU06pnOldArray): IAU 2006 data (old nutation coefficients)

    Returns:
        tuple:
            deltapsi (float): Change in longitude in radians
            pnb (np.ndarray): Combined precession-nutation matrix
            prec (np.ndarray): Precession transformation matrix (MOD to J2000)
            nut (np.ndarray): Nutation transformation matrix (IRE to GCRF)
            fundargs (FundArgs): Delaunay and planetary arguments
    """
    # Definitions
    iau2000b_terms = 77

    # Obtain data for calculations from the 2000b theory
    fundargs = fundarg(ttt, opt="02")

    # Compute luni-solar nutation
    pnsum, ensum = 0, 0
    for i in range(iau2000b_terms - 1, -1, -1):
        tempval = (
            iau06arr.apni[i, 0] * fundargs.l
            + iau06arr.apni[i, 1] * fundargs.l1
            + iau06arr.apni[i, 2] * fundargs.f
            + iau06arr.apni[i, 3] * fundargs.d
            + iau06arr.apni[i, 4] * fundargs.omega
        )
        pnsum += (iau06arr.apn[i, 0] + iau06arr.apn[i, 1] * ttt) * np.sin(tempval) + (
            iau06arr.apn[i, 4] + iau06arr.apn[i, 5] * ttt
        ) * np.cos(tempval)
        ensum += (iau06arr.apn[i, 2] + iau06arr.apn[i, 3] * ttt) * np.cos(tempval) + (
            iau06arr.apn[i, 6] + iau06arr.apn[i, 7] * ttt
        ) * np.sin(tempval)

    # Planetary nutation constants
    pplnsum = -0.000135 * const.ARCSEC2RAD
    eplnsum = 0.000388 * const.ARCSEC2RAD

    # Combine nutation components
    deltapsi = pnsum + pplnsum
    deltaeps = ensum + eplnsum

    # Build transformation matrices
    nut, prec, pnb = _build_transformation_matrices(ttt, deltaeps, deltapsi, True)

    return deltapsi, pnb, prec, nut, fundargs


########################################################################################
# IAU 2006 XYS Parameters
########################################################################################


def _get_mfme_recnum(jd: float, jdf: float, mjd0: float) -> Tuple[float, int]:
    # Calculate the Julian day at 0000 hr
    jdb = np.floor(jd + jdf) + 0.5
    mfme = (jd + jdf - jdb) * const.DAY2MIN
    if mfme < 0:
        mfme += const.DAY2MIN

    # Find the record number corresponding to the desired day
    recnum = int(np.floor(jd + jdf - mjd0 - const.JD_TO_MJD_OFFSET))

    return mfme, recnum


def findxysparam(
    jdtt: float,
    jdttf: float,
    iau06xysarr: IAU06xysArray,
    interp: InterpolationMode | None = None,
) -> Tuple[float, float, float]:
    """Finds the X, Y, S parameters for a given time with optional interpolation.

    References:
        Vallado, 2013

    Args:
        jdtt (float): Epoch Julian day (days from 4713 BC)
        jdttf (float): Epoch Julian day fraction (day fraction from jdutc)
        iau06xysarr (IAU06xysArray): IAU 2006 XYS data
        interp (InterpolationMode, optional): Interpolation mode (default: None)

    Returns:
        tuple: (x, y, s)
            x (float): X component of CIO in radians
            y (float): Y component of CIO in radians
            s (float): S component in radians
    """
    # Find the record number corresponding to the desired day
    mfme, recnum = _get_mfme_recnum(jdtt, jdttf, iau06xysarr.mjd_tt[0])

    # Check for out-of-bound values
    if 0 <= recnum <= len(iau06xysarr.x) - 1:
        if interp == InterpolationMode.LINEAR:
            # Perform linear interpolation
            target_time = iau06xysarr.mjd_tt[recnum] + mfme / const.DAY2MIN
            x = np.interp(
                target_time,
                iau06xysarr.mjd_tt[recnum : recnum + 2],
                iau06xysarr.x[recnum : recnum + 2],
            )
            y = np.interp(
                target_time,
                iau06xysarr.mjd_tt[recnum : recnum + 2],
                iau06xysarr.y[recnum : recnum + 2],
            )
            s = np.interp(
                target_time,
                iau06xysarr.mjd_tt[recnum : recnum + 2],
                iau06xysarr.s[recnum : recnum + 2],
            )
        elif interp == InterpolationMode.SPLINE:
            # Perform cubic spline interpolation
            start_idx = max(0, recnum - 1)
            end_idx = min(len(iau06xysarr.x), recnum + 3)
            cs_x = CubicSpline(
                iau06xysarr.mjd_tt[start_idx:end_idx], iau06xysarr.x[start_idx:end_idx]
            )
            cs_y = CubicSpline(
                iau06xysarr.mjd_tt[start_idx:end_idx], iau06xysarr.y[start_idx:end_idx]
            )
            cs_s = CubicSpline(
                iau06xysarr.mjd_tt[start_idx:end_idx], iau06xysarr.s[start_idx:end_idx]
            )
            target_time = iau06xysarr.mjd_tt[recnum] + mfme / const.DAY2MIN
            x = cs_x(target_time).item()
            y = cs_y(target_time).item()
            s = cs_s(target_time).item()
        else:
            # No interpolation
            x = iau06xysarr.x[recnum]
            y = iau06xysarr.y[recnum]
            s = iau06xysarr.s[recnum]
    else:
        # Default values for out-of-bound requests
        x, y, s = 0, 0, 0

    return x, y, s


def create_xys(
    iau06arr: IAU06Array,
    directory: str | None = None,
    filename: str = "xysdata.dat",
    yr_span: int = 1,
    dt_day: int = 1,
    ymdhms: Tuple[int, int, int, int, int, float] = (1957, 1, 1, 0, 0, 0.0),
) -> np.ndarray:
    """Generate the XYS data array and optionally save to a file.

    This function precalculates the XYS parameters and optionally stores in a data file
    for efficient access in the future.

    Args:
        iau06arr (IAU06Array): IAU 2006 data
        directory (str, optional): Directory to save the output file (default: None)
        filename (str, optional): Output filename (default: 'xysdata.dat')
        yr_span (int, optional): Number of years to generate data (default: 142 years)
        dt_day (int, optional): Time step in days (default: 1 day)
        ymdhms (tuple, optional): Initial date in (yr, mo, day, hr, min, sec) format
                                  (default: (1957, 1, 1, 0, 0, 0.0))

    Returns:
        np.ndarray: Array of XYS data with columns [jdtt, jdftt, x, y, s]

    Notes:
        - This is pretty slow due to `iau06xys_series` being called for each day and
          could use some optimization.
        - MATLAB and C# versions hardcode the year start and duration; the default
          duration is set to 1 year here instead of the 142 years used in those

    TODO:
        - Look into using `jit` for performance improvements of downstream functions
    """
    # Initialize the starting Julian date
    jdtt, jdftt = jday(*ymdhms)

    # Calculate the number of rows for the array
    num_rows = yr_span * int(const.YR2DAY) // dt_day + 1

    # Pre-initialize the data array
    xys_data = np.zeros((num_rows, 5))

    # Generate the data
    for i in range(num_rows):
        ttt = (jdtt + jdftt - const.J2000) / const.CENT2DAY
        x, y, s = iau06xys_series(ttt, iau06arr)

        # Store data in the array
        xys_data[i] = [jdtt, jdftt, x, y, s]

        # Increment the Julian date
        jdtt += dt_day

    # Optionally save to a file
    if directory:
        np.savetxt(
            f"{directory}/{filename}",
            xys_data,
            fmt=["%15.6f", "%13.11f", "%15.12f", "%15.12f", "%15.12f"],
            header="jdtt jdftt x y s",
            comments="",
        )

    return xys_data


def iau06xys_series(ttt: float, iau06arr: IAU06Array) -> Tuple[float, float, float]:
    """Calculates the XYS parameters for the IAU2006 CIO theory.

    This is the series implementation of the XYS parameters, which are used to compute
    the Celestial Intermediate Origin (CIO) locator.

    References:
        Vallado, 2022, p. 214-216

    Args:
        ttt (float): Julian centuries of TT
        iau06arr (IAU06Array): IAU 2006 data

    Returns:
        tuple: (x, y, s)
            x (float): Coordinate of CIP in radians
            y (float): Coordinate of CIP in radians
            s (float): Coordinate in radians
    """
    # Fundamental arguments from the IAU 2006 nutation theory
    fundargs = fundarg(ttt, opt="06")

    # Powers of TTT
    ttt2, ttt3, ttt4, ttt5 = ttt**2, ttt**3, ttt**4, ttt**5

    # Limits for the x, y, and s series. These numbers correspond to the ranges of
    # terms used in the calculations for each group:
    # - Group 1: Main series (1306 terms for x, 962 for y, etc.)
    # - Group 2: Secondary contributions (253 terms for x, 277 for y, etc.)
    # - Group 3: Smaller corrections (36 terms for x, 30 for y, etc.)
    # - Group 4: Even smaller corrections (4 terms for x, 5 for y, etc.)
    # - Group 5: Minimal corrections (1 term each for both x and y, 0 for s)

    # Compute X
    limits_x = [1306, 253, 36, 4, 1]  # total sum = 1600 (axs0 and a0xi length)
    x_sums = [0] * len(limits_x)

    # Loop over each group
    for group, limit in enumerate(limits_x):
        start_index = sum(limits_x[:group])
        for i in range(limit):
            idx = start_index + i
            tempval = (
                iau06arr.ax0i[idx, 0] * fundargs.l
                + iau06arr.ax0i[idx, 1] * fundargs.l1
                + iau06arr.ax0i[idx, 2] * fundargs.f
                + iau06arr.ax0i[idx, 3] * fundargs.d
                + iau06arr.ax0i[idx, 4] * fundargs.omega
                + iau06arr.ax0i[idx, 5] * fundargs.lonmer
                + iau06arr.ax0i[idx, 6] * fundargs.lonven
                + iau06arr.ax0i[idx, 7] * fundargs.lonear
                + iau06arr.ax0i[idx, 8] * fundargs.lonmar
                + iau06arr.ax0i[idx, 9] * fundargs.lonjup
                + iau06arr.ax0i[idx, 10] * fundargs.lonsat
                + iau06arr.ax0i[idx, 11] * fundargs.lonurn
                + iau06arr.ax0i[idx, 12] * fundargs.lonnep
                + iau06arr.ax0i[idx, 13] * fundargs.precrate
            )
            x_sums[group] += iau06arr.ax0[idx, 0] * np.sin(tempval) + iau06arr.ax0[
                idx, 1
            ] * np.cos(tempval)

    # Final value for x
    x = (
        -0.016617
        + 2004.191898 * ttt
        - 0.4297829 * ttt2
        - 0.19861834 * ttt3
        - 0.000007578 * ttt4
        + 0.0000059285 * ttt5
    )
    x = x * const.ARCSEC2RAD + sum(x_sums * np.array([1, ttt, ttt2, ttt3, ttt4]))

    # Compute Y
    limits_y = [962, 277, 30, 5, 1]  # total sum = 1275 (ays0 and a0yi length)
    y_sums = [0] * len(limits_y)

    # Loop over each group
    for group, limit in enumerate(limits_y):
        start_index = sum(limits_y[:group])
        for i in range(limit):
            idx = start_index + i
            tempval = (
                iau06arr.ay0i[idx, 0] * fundargs.l
                + iau06arr.ay0i[idx, 1] * fundargs.l1
                + iau06arr.ay0i[idx, 2] * fundargs.f
                + iau06arr.ay0i[idx, 3] * fundargs.d
                + iau06arr.ay0i[idx, 4] * fundargs.omega
                + iau06arr.ay0i[idx, 5] * fundargs.lonmer
                + iau06arr.ay0i[idx, 6] * fundargs.lonven
                + iau06arr.ay0i[idx, 7] * fundargs.lonear
                + iau06arr.ay0i[idx, 8] * fundargs.lonmar
                + iau06arr.ay0i[idx, 9] * fundargs.lonjup
                + iau06arr.ay0i[idx, 10] * fundargs.lonsat
                + iau06arr.ay0i[idx, 11] * fundargs.lonurn
                + iau06arr.ay0i[idx, 12] * fundargs.lonnep
                + iau06arr.ay0i[idx, 13] * fundargs.precrate
            )
            y_sums[group] += iau06arr.ay0[idx, 0] * np.sin(tempval) + iau06arr.ay0[
                idx, 1
            ] * np.cos(tempval)

    # Final value for y
    y = (
        -0.006951
        - 0.025896 * ttt
        - 22.4072747 * ttt2
        + 0.00190059 * ttt3
        + 0.001112526 * ttt4
        + 0.0000001358 * ttt5
    )
    y = y * const.ARCSEC2RAD + sum(y_sums * np.array([1, ttt, ttt2, ttt3, ttt4]))

    # Compute S
    limits_s = [33, 3, 25, 4, 1]  # total sum = 66 (ass0 and a0si length)
    s_sums = [0] * len(limits_s)

    # Loop over each group
    for group, limit in enumerate(limits_s):
        start_index = sum(limits_s[:group])
        for i in range(limit):
            idx = start_index + i
            tempval = (
                iau06arr.as0i[idx, 0] * fundargs.l
                + iau06arr.as0i[idx, 1] * fundargs.l1
                + iau06arr.as0i[idx, 2] * fundargs.f
                + iau06arr.as0i[idx, 3] * fundargs.d
                + iau06arr.as0i[idx, 4] * fundargs.omega
                + iau06arr.as0i[idx, 5] * fundargs.lonmer
                + iau06arr.as0i[idx, 6] * fundargs.lonven
                + iau06arr.as0i[idx, 7] * fundargs.lonear
                + iau06arr.as0i[idx, 8] * fundargs.lonmar
                + iau06arr.as0i[idx, 9] * fundargs.lonjup
                + iau06arr.as0i[idx, 10] * fundargs.lonsat
                + iau06arr.as0i[idx, 11] * fundargs.lonurn
                + iau06arr.as0i[idx, 12] * fundargs.lonnep
                + iau06arr.as0i[idx, 13] * fundargs.precrate
            )
            s_sums[group] += iau06arr.as0[idx, 0] * np.sin(tempval) + iau06arr.as0[
                idx, 1
            ] * np.cos(tempval)

    # Final value for s
    s = (
        0.000094
        + 0.00380865 * ttt
        - 0.00012268 * ttt2
        - 0.07257411 * ttt3
        + 0.00002798 * ttt4
        + 0.00001562 * ttt5
    )
    s = (
        -x * y * 0.5
        + s * const.ARCSEC2RAD
        + sum(s_sums * np.array([1, ttt, ttt2, ttt3, ttt4]))
    )

    return x, y, s


def iau06xys(
    ttt: float,
    iau06arr: IAU06Array,
    ddx: float = 0.0,
    ddy: float = 0.0,
    iau06xysarr: IAU06xysArray | None = None,
    use_full_series: bool = True,
) -> Tuple[float, float, float, np.ndarray]:
    """Calculates the transformation matrix that accounts for the effects of
    precession-nutation using the IAU2006 theory.

    References:
        Vallado, 2022, pp. 214, 221

    Args:
        ttt (float): Julian centuries of TT
        iau06arr (IAU06Array): IAU 2006 data
        ddx (float, optional): EOP correction for x in radians (default is 0)
        ddy (float, optional): EOP correction for y in radians (default is 0)
        iau06xysarr (IAU06xysArray, optional): IAU 2006 XYS data (default is None)
        use_full_series (bool, optional): Whether to use the full series implementation
                                          for XYS parameters (default is True)

    Returns:
        tuple: (x, y, s, pn)
            x (float): Coordinate of CIP in radians
            y (float): Coordinate of CIP in radians
            s (float): Coordinate in radians
            pn (np.ndarray): Transformation matrix for TIRS-GCRF
    """
    # Calculate X, Y, and S components
    if use_full_series:
        # Use the full series implementation
        x, y, s = iau06xys_series(ttt, iau06arr)
    else:
        # Check that the XYS array is provided
        if not iau06xysarr:
            raise ValueError("IAU 2006 XYS array must be provided for interpolation!")

        # Find the X, Y, and S components using spline interpolation
        # TODO: allow for user to specify interpolation method?
        jdtt = ttt * const.CENT2DAY + const.J2000
        x, y, s = findxysparam(jdtt, 0, iau06xysarr, InterpolationMode.SPLINE)

    # Apply any corrections for x and y
    x += ddx
    y += ddy

    # Calculate the 'a' parameter based on x and y
    a = 0.5 + 0.125 * (x**2 + y**2)

    # Build nutation matrices
    nut1 = np.array(
        [
            [1 - a * x**2, -a * x * y, x],
            [-a * x * y, 1 - a * y**2, y],
            [-x, -y, 1 - a * (x**2 + y**2)],
        ]
    )
    nut2 = np.array([[np.cos(s), np.sin(s), 0], [-np.sin(s), np.cos(s), 0], [0, 0, 1]])

    # Combine to form the final transformation matrix
    pn = np.dot(nut1, nut2)

    return x, y, s, pn


########################################################################################
# Earth Orientation Parameters (EOP)
########################################################################################


def findeopparam(
    jd: float, jdf: float, eoparr: EOPArray, interp: InterpolationMode | None = None
) -> Tuple[float, float, float, float, float, float, float, float, float]:
    """Finds the EOP parameters for a given time with optional interpolation.

    References:
        Vallado, 2013

    Args:
        jd (float): Epoch Julian day (days from 4713 BC)
        jdf (float): Epoch Julian day fraction (day fraction from jdutc)
        eoparr (EOPArray): EOP data
        interp (InterpolationMode, optional): Interpolation mode (default: None)

    Returns:
        tuple: (dut1, dat, lod, xp, yp, ddpsi, ddeps, dx, dy)
            dut1 (float): Julian date of UT1 (days from 4713 BC)
            dat (int): TAI - UTC in seconds
            lod (float): Length of day in seconds
            xp (float): Polar motion coefficient in radians
            yp (float): Polar motion coefficient in radians
            ddpsi (float): Delta psi (nutation in longitude) correction in radians
            ddeps (float): Delta epsilon (nutation in obliquity) correction in radians
            dx (float): Celestial pole (CIP) x offset in radians
            dy (float): Celestial pole (CIP) y offset in radians
    """
    # Find the record number corresponding to the desired day
    mfme, recnum = _get_mfme_recnum(jd, jdf, eoparr.mjd[0])

    # Ensure recnum is within valid bounds
    if 0 <= recnum < len(eoparr.mjd) - 1:
        mjd = eoparr.mjd
        params = np.vstack(
            [
                eoparr.dut1,
                eoparr.dat,
                eoparr.lod,
                eoparr.xp,
                eoparr.yp,
                eoparr.ddpsi,
                eoparr.ddeps,
                eoparr.dx,
                eoparr.dy,
            ]
        ).T

        if interp == InterpolationMode.LINEAR:
            # Linear interpolation
            fixf = mfme / const.DAY2MIN
            weights = [1 - fixf, fixf]
            interp_params = params[recnum : recnum + 2].T @ weights
        elif interp == InterpolationMode.SPLINE:
            # Cubic spline interpolation
            cs = CubicSpline(
                mjd[recnum - 1 : recnum + 3], params[recnum - 1 : recnum + 3], axis=0
            )
            interp_params = cs(mjd[recnum] + mfme / const.DAY2MIN)
        else:
            # No interpolation
            interp_params = params[recnum]
    else:
        # Default values for out-of-bounds requests
        interp_params = np.zeros(9)

    # Convert units for certain parameters
    interp_params[1] = int(interp_params[1])
    interp_params[3:9] *= const.ARCSEC2RAD

    return tuple(interp_params)



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 1 March 2001
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------

from typing import Tuple

import numpy as np
from numpy.typing import ArrayLike

from ...constants import RE, MU


def hillsr(
    r: ArrayLike, v: ArrayLike, alt: float, dts: float
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculate position and velocity information for Hill's (Clohessy-Wiltshire)
    equations.

    References:
        Vallado: 2022, p. 397-401, Algorithm 48

    Args:
        r (array_like): Initial relative position of the interceptor in km
        v (array_like): Initial relative velocity of the interceptor in km/s
        alt (float): Altitude of the target satellite in km
        dts (float): Desired time in seconds

    Returns:
        tuple: (rint, vint)
            rint (np.ndarray): Final relative position of the interceptor in km
            vint (np.ndarray): Final relative velocity of the interceptor in km/s

    Notes:
        - Position and velocity vectors are in the RSW frame
        - Distance units for r and v are flexible, but must be consistent
    """
    # Calculate orbital parameters
    radius = RE + alt
    omega = np.sqrt(MU / (radius**3))
    nt = omega * dts
    cosnt = np.cos(nt)
    sinnt = np.sin(nt)

    # Determine new positions
    rint = np.zeros(3)
    rint[0] = (
        (v[0] / omega) * sinnt
        - ((2 * v[1] / omega) + 3 * r[0]) * cosnt
        + ((2 * v[1] / omega) + 4 * r[0])
    )
    rint[1] = (
        (2 * v[0] / omega) * cosnt
        + ((4 * v[1] / omega) + 6 * r[0]) * sinnt
        + (r[1] - (2 * v[0] / omega))
        - (3 * v[1] + 6 * omega * r[0]) * dts
    )
    rint[2] = r[2] * cosnt + (v[2] / omega) * sinnt

    # Determine new velocities
    vint = np.zeros(3)
    vint[0] = v[0] * cosnt + (2 * v[1] + 3 * omega * r[0]) * sinnt
    vint[1] = (
        -2 * v[0] * sinnt
        + (4 * v[1] + 6 * omega * r[0]) * cosnt
        - (3 * v[1] + 6 * omega * r[0])
    )
    vint[2] = -r[2] * omega * sinnt + v[2] * cosnt

    return rint, vint


def hillsv(r: ArrayLike, alt: float, dts: float, tol: float = 1e-6) -> np.ndarray:
    """Calculate the initial velocity for Hill's (Clohessy-Wiltshire) equations.

    References:
        Vallado: 2022, p. 410-414, Eq. 6-66

    Args:
        r (array_like): Initial position vector of the interceptor in km
        alt (float): Altitude of the target satellite in km
        dts (float): Desired time in seconds
        tol (float, optional): Tolerance for calculations (defaults to 1e-6)

    Returns:
        np.ndarray: Initial velocity vector of the interceptor in km/s

    Notes:
        - Position and velocity vectors are in the RSW frame
        - Distance units for r are flexible, and velocity units are consistent
    """
    # Calculate the orbital parameters
    radius = RE + alt
    omega = np.sqrt(MU / (radius**3))
    nt = omega * dts
    cosnt, sinnt = np.cos(nt), np.sin(nt)

    # Numerator and denominator for the initial velocity
    numkm = (6 * r[0] * (nt - sinnt) - r[1]) * omega * sinnt - 2 * omega * r[0] * (
        4 - 3 * cosnt
    ) * (1 - cosnt)
    denom = (4 * sinnt - 3 * nt) * sinnt + 4 * (1 - cosnt) ** 2

    # Determine initial velocity
    v = np.zeros(3)
    v[1] = numkm / denom if abs(denom) > tol else 0
    if abs(sinnt) > tol:
        v[0] = -(omega * r[0] * (4 - 3 * cosnt) + 2 * (1 - cosnt) * v[1]) / sinnt
    v[2] = -r[2] * omega / np.tan(nt)

    return v



# --------------------------------------------------------------------------------------
# Author: David Vallado
# Date: 7 June 2002
#
# Copyright (c) 2024
# For license information, see LICENSE file
# --------------------------------------------------------------------------------------


import numpy as np
from typing import Tuple

from ... import constants as const
from ...mathtime.julian_date import jday


def gstime(jdut1: float) -> float:
    """Calculates the Greenwich Sidereal Time (IAU-82).

    References:
        Vallado: 2022, p. 189, Eq. 3-48

    Args:
        jdut1 (float): Julian date of UT1 (days from 4713 BC)

    Returns:
        float: Greenwich Sidereal Time in radians (0 to 2pi)
    """
    # Julian centuries from the J2000 epoch
    tut1 = (jdut1 - const.J2000) / const.CENT2DAY

    # Calculate Greenwich Sidereal Time in seconds
    gst = (
        -6.2e-6 * tut1**3
        + 0.093104 * tut1**2
        + (876600 * const.HR2SEC + 8640184.812866) * tut1
        + 67310.54841
    )

    # Convert to radians
    return np.remainder(gst * const.EARTHROT_APPROX, const.TWOPI)


def gstime0(year: int) -> float:
    """Calculates the Greenwich Sidereal Time at the beginning (0 hr UT1 on January 1)
    of the given year.

    References:
        Vallado: 2022, p. 189-190, Eq. 3-48

    Args:
        year (int): Year (e.g., 1998, 1999, etc.)

    Returns:
        float: Greenwich Sidereal Time in radians (0 to 2pi)
    """
    # Calculate Julian Date at 0 hr UT1 on January 1
    jd, _ = jday(year, month=1, day=1)

    return gstime(jd)


def lstime(lon: float, jdut1: float) -> Tuple[float, float]:
    """Calculates the local sidereal time (LST) and Greenwich sidereal time (GST) at a
    given location (GST from IAU-82).

    References:
        Vallado: 2022, p. 190, Algorithm 15

    Args:
        lon (float): Longitude of the site in radians (-2pi to 2pi) (West is negative)
        jdut1 (float): Julian date of UT1 (days from 4713 BC)

    Returns:
        tuple: (lst, gst)
            lst (float): Local sidereal time (LST) in radians (0 to 2pi)
            gst (float): Greenwich sidereal time (GST) in radians (0 to 2pi)
    """
    # Calculate GST
    gst = gstime(jdut1)

    # Calculate LST
    lst = np.mod(lon + gst, const.TWOPI)

    return lst, gst


def sidereal(
    jdut1: float,
    deltapsi: float,
    meaneps: float,
    omega: float,
    lod: float,
    use_iau80: bool = True,
    eqeterms: bool = True,
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculates the transformation matrix that accounts for the effects of
    sidereal time.

    References:
        Vallado: 2022, p. 224-225

    Args:
        jdut1 (float): Julian date of UT1 (days from 4713 BC)
        deltapsi (float): Nutation angle in radians
        meaneps (float): Mean obliquity of the ecliptic in radians
        omega (float): Longitude of ascending node of the moon in radians
        lod (float): Length of day in seconds
        use_iau80 (bool, optional): Use IAU-80 theory (default True)
        eqeterms (bool, optional): Add terms for ast calculation (default True)

    Returns:
        tuple: (st, stdot)
            st (np.ndarray): Transformation matrix for PEF to TOD
            stdot (np.ndarray): Transformation rate matrix
    """
    # Calculate apparent GMST
    if use_iau80:
        # Find GMST
        gmst = gstime(jdut1)

        # Find mean apparent sidereal time
        if jdut1 > 2450449.5 and eqeterms:
            ast = (
                gmst
                + deltapsi * np.cos(meaneps)
                + 0.00264 * const.ARCSEC2RAD * np.sin(omega)
                + 0.000063 * const.ARCSEC2RAD * np.sin(2 * omega)
            )
        else:
            ast = gmst + deltapsi * np.cos(meaneps)

        ast = np.remainder(ast, const.TWOPI)
    else:
        tut1d = jdut1 - const.J2000
        era = const.TWOPI * (0.7790572732640 + 1.00273781191135448 * tut1d)
        ast = np.remainder(era, const.TWOPI)

    # Transformation matrix for PEF to TOD
    st = np.array(
        [[np.cos(ast), -np.sin(ast), 0], [np.sin(ast), np.cos(ast), 0], [0, 0, 1]]
    )

    # Sidereal time rate matrix
    omegaearth = const.EARTHROT * (1 - lod / const.DAY2SEC)
    stdot = np.array(
        [
            [-omegaearth * np.sin(ast), -omegaearth * np.cos(ast), 0],
            [omegaearth * np.cos(ast), -omegaearth * np.sin(ast), 0],
            [0, 0, 0],
        ]
    )

    return st, stdot
